import click
from hamlet.backend.manage import credential_crypto as manage_credentials_crypto_backend
from hamlet.command.common.config import pass_options
from hamlet.command.common import exceptions


@click.command(
    "credential-crypto",
    short_help="Manage crypto for credential storage",
    context_settings=dict(max_content_width=240),
    deprecated=True,
)
@click.option(
    "-e",
    "--credential-email",
    help="email associated with the credential (not encrypted)",
)
@click.option(
    "-f",
    "--crypto-file",
    help="path to the credentials file to be used",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
)
@click.option(
    "-i",
    "--credential-id",
    help="id of credential (i.e. Username/Client Key/Access Key value) - not encrypted",
)
@click.option(
    "-n",
    "--credential-path",
    help="path for the set of values (id, secret, email)",
    required=True,
)
@click.option(
    "-s",
    "--credential-secret",
    help="secret of credential (i.e. Password/Secret Key value) - encrypted",
)
@click.option(
    "-v",
    "--visible",
    help="if credential secret should be decrypted (visible)",
    is_flag=True,
)
@click.option(
    "-y",
    "--credential-type",
    help="type of credential",
    required=True,
    type=click.Choice(["login", "api", "env"], case_sensitive=False),
)
@exceptions.backend_handler()
@pass_options
def credentials_crypto(options, **kwargs):
    """
    Manage crypto for credential storage

    \b
    NOTES:
    1. CREDENTIAL_PATH is a JSON path so values separated by dots. It is case sensitive.
    2. CREDENTIAL_TYPE is ${CREDENTIAL_TYPE_LOGIN}, ${CREDENTIAL_TYPE_API} or ${CREDENTIAL_TYPE_ENV}
    3. For CREDENTIAL_TYPE of ${CREDENTIAL_TYPE_LOGIN}, Id Attribute = Username, Secret Attribute = Password
    4. For CREDENTIAL_TYPE of ${CREDENTIAL_TYPE_API}, Id Attribute = AccessKey, Secret Attribute = SecretKey
    5. For CREDENTIAL_TYPE of ${CREDENTIAL_TYPE_ENV}, Id Attribute = ACCESS_KEY, Secret Attribute = SECRET_KEY
    """

    args = {**options.opts, **kwargs}

    manage_credentials_crypto_backend.run(**args, engine=options.engine, _is_cli=True)



from hamlet.backend.common import runner


def run(
    deployment_unit=None,
    deployment_group=None,
    sentry_source_map_s3_url=None,
    sentry_url_prefix=None,
    sentry_release_name=None,
    app_type=None,
    dist=None,
    log_level=None,
    log_format=None,
    root_dir=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    engine=None,
    _is_cli=False,
    **kwargs
):
    options = {
        "-m": sentry_source_map_s3_url,
        "-p": sentry_url_prefix,
        "-r": sentry_release_name,
        "-u": deployment_unit,
        "-g": deployment_group,
        "-a": app_type,
        "-d": dist,
    }
    env = {
        "GENERATION_LOG_LEVEL": log_level,
        "GENERATION_LOG_FORMAT": log_format,
        "ROOT_DIR": root_dir,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
    }
    runner.run(
        "runSentryRelease.sh",
        args=[],
        options=options,
        engine=engine,
        env=env,
        _is_cli=_is_cli,
    )



import os
import click


from tabulate import tabulate

from hamlet.command import root as cli
from hamlet.command.common.display import json_or_table_option, wrap_text
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options
from hamlet.command.common.validate import validate_entrance_inputs
from hamlet.backend.create import template as create_template_backend
from hamlet.backend import query as query_backend


@cli.group("entrance", context_settings=dict(max_content_width=240))
def group():
    """
    Hamlet entrances provide access to the hamlet cmdb to perform different tasks
    """
    pass


LIST_ENTRANCES_QUERY = "Entrances[].{Type:Type,Description:Description}"


def entrances_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["Type"]),
                wrap_text(row["Description"]),
            ]
        )
    return tabulate(
        tablerows,
        headers=["Type", "Description"],
        showindex=True,
        tablefmt="fancy_grid",
    )


@group.command(
    "list-entrances", short_help="", context_settings=dict(max_content_width=240)
)
@json_or_table_option(entrances_table)
@exceptions.backend_handler()
@pass_options
def list_entrances(options):
    """
    List available entrances
    """
    args = {
        **options.opts,
        "generation_input_source": "mock",
        "generation_entrance": "info",
        "output_filename": "info.json",
    }

    return query_backend.run(
        **args, cwd=os.getcwd(), query_text=LIST_ENTRANCES_QUERY, engine=options.engine
    )


@group.command(
    "invoke-entrance", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-e",
    "--entrance",
    help="The entrance to invoke for output generation",
    required=True,
)
@click.option(
    "-y",
    "--entrance-parameter",
    multiple=True,
    help="key=value pairs that are passed to to the entrance",
    callback=validate_entrance_inputs,
)
@click.option(
    "-l",
    "--deployment-group",
    help="deployment group the deployment unit belongs to",
)
@click.option(
    "-u",
    "--deployment-unit",
    help="deployment unit to be included in the template",
)
@click.option(
    "-z", "--deployment-unit-subset", help="subset of the deployment unit required"
)
@click.option(
    "-d",
    "--deployment-mode",
    help="deployment mode the template will be generated for",
    default="update",
    show_default=True,
)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
    ),
    help="the directory where the outputs will be saved. Mandatory when input source is set to mock",
)
@click.option(
    "-x",
    "--disable-output-cleanup",
    is_flag=True,
    help="disable the cleanup of the output directory before generation",
)
@click.option(
    "-c",
    "--config-ref",
    help="identifier of the configuration used to generate this template",
    default="unassigned",
    show_default=True,
)
@click.option(
    "-q",
    "--request-ref",
    help="opaque value to link this template to a triggering request management system",
    default="unassigned",
    show_default=True,
)
@exceptions.backend_handler()
@pass_options
def invoke_entrance(options, **kwargs):
    """
    Invoke a Hamlet Entrance
    """
    args = {**options.opts, **kwargs}

    create_template_backend.run(**args, engine=options.engine, _is_cli=True)



import os
import json
from unittest import mock
from click.testing import CliRunner

from hamlet.command.component.describe_occurrence import describe_occurrence
from hamlet.command.component.list_occurrences import list_occurrences
from hamlet.command.component.query_occurrences import query_occurrences

from hamlet.command.component.common import DescribeContext


def template_backend_run_mock(data):
    def run(output_filename="occurrences-state.json", output_dir=None, *args, **kwargs):
        filename = os.path.join(output_dir, output_filename)
        with open(filename, "wt+") as f:
            json.dump(data, f)

    return run


def mock_backend(data=None):
    def decorator(func):
        @mock.patch("hamlet.backend.query.template")
        def wrapper(blueprint_mock, *args, **kwargs):

            blueprint_mock.run.side_effect = template_backend_run_mock(data)
            return func(blueprint_mock, *args, **kwargs)

        return wrapper

    return decorator


def mock_describe_context(name=None):
    def decorator(func):
        @mock.patch.object(DescribeContext, "name", name)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return decorator


occurrence_state_data = {
    "Occurrences": [
        {
            "Core": {
                "Tier": {"Id": "TierId[1]"},
                "Component": {"Id": "ComponentId[1]", "RawId": "ComponentRawId[1]"},
                "TypedRawName": "CoreRawName[1]",
                "Type": "CoreType[1]",
            },
            "Configuration": {
                "Solution": {"deployment:Unit": "DeploymentUnit[1]"},
                "SettingNamespaces": [{"Key": "SettingNamespace[1]"}],
            },
            "State": {
                "Resources": {
                    "Resource[1]": {"Id": "ResourceId[1]", "Name": "ResourceName[1]"}
                },
                "Attributes": {"ATTRIBUTE[1]": "AttributeValue[1]"},
            },
        },
        {
            "Core": {
                "Tier": {"Id": "TierId[2]"},
                "Component": {"Id": "ComponentId[2]", "RawId": "ComponentRawId[2]"},
                "TypedRawName": "CoreRawName[2]",
                "Type": "CoreType[2]",
            },
            "Configuration": {
                "Solution": {"deployment:Unit": "DeploymentUnit[2]"},
                "SettingNamespaces": [{"Key": "SettingNamespace[2]"}],
            },
            "State": {
                "Resources": {
                    "Resource[1]": {"Id": "ResourceId[2]", "Name": "ResourceName[2]"}
                },
                "Attributes": {"ATTRIBUTE[1]": "AttributeValue[2]"},
            },
        },
    ]
}


@mock_backend(occurrence_state_data)
def test_list_occurrences(blueprint_mock):
    cli = CliRunner()
    result = cli.invoke(list_occurrences, ["--output-format", "json"])
    print(result.stdout)
    assert result.exit_code == 0
    result = json.loads(result.output)
    print(result)
    assert len(result) == 2
    assert {
        "TierId": "TierId[1]",
        "ComponentId": "ComponentRawId[1]",
        "Name": "CoreRawName[1]",
        "Type": "CoreType[1]",
    } in result
    assert {
        "TierId": "TierId[2]",
        "ComponentId": "ComponentRawId[2]",
        "Name": "CoreRawName[2]",
        "Type": "CoreType[2]",
    } in result


@mock_describe_context("CoreRawName[1]")
@mock_backend(occurrence_state_data)
def test_describe_occurrence(blueprint_mock):

    cli = CliRunner()
    result = cli.invoke(describe_occurrence, ["--name", "CoreRawName[1]"])
    print(result.exc_info)
    assert result.exit_code == 0, result.output
    assert json.loads(result.output) == occurrence_state_data["Occurrences"][0]


@mock_describe_context("CoreRawName[1]")
@mock_backend(occurrence_state_data)
def test_describe_occurrence_query(blueprint_mock):

    cli = CliRunner()
    result = cli.invoke(
        describe_occurrence,
        ["--name", "CoreRawName[1]", "--query", "Configuration.Solution"],
    )
    print(result.exc_info)
    assert result.exit_code == 0, result.output
    assert json.loads(result.output) == {"deployment:Unit": "DeploymentUnit[1]"}


@mock_describe_context("CoreRawName[1]")
@mock_backend(occurrence_state_data)
def test_describe_occurrence_query_solution(blueprint_mock):

    cli = CliRunner()
    result = cli.invoke(describe_occurrence, ["--name", "CoreRawName[1]", "solution"])
    print(result.exc_info)
    assert result.exit_code == 0, result.output
    assert json.loads(result.output) == {"deployment:Unit": "DeploymentUnit[1]"}


@mock_describe_context("CoreRawName[1]")
@mock_backend(occurrence_state_data)
def test_describe_occurrence_query_resources(blueprint_mock):

    cli = CliRunner()
    result = cli.invoke(describe_occurrence, ["--name", "CoreRawName[1]", "resources"])
    print(result.exc_info)
    assert result.exit_code == 0, result.output
    assert json.loads(result.output) == {
        "Resource[1]": {"Id": "ResourceId[1]", "Name": "ResourceName[1]"}
    }


@mock_describe_context("CoreRawName[1]")
@mock_backend(occurrence_state_data)
def test_describe_occurrence_query_setting_namespaces(blueprint_mock):

    cli = CliRunner()
    result = cli.invoke(
        describe_occurrence, ["--name", "CoreRawName[1]", "setting-namespaces"]
    )
    print(result.exc_info)
    assert result.exit_code == 0, result.output
    assert json.loads(result.output) == [{"Key": "SettingNamespace[1]"}]


@mock_describe_context("CoreRawName[1]")
@mock_backend(occurrence_state_data)
def test_describe_occurrence_query_attributes(blueprint_mock):

    cli = CliRunner()
    result = cli.invoke(
        describe_occurrence,
        [
            "--name",
            "CoreRawName[1]",
            "attributes",
            "--output-format",
            "json",
        ],
    )
    print(result.exc_info)
    assert result.exit_code == 0, result.output
    assert json.loads(result.output) == {"ATTRIBUTE[1]": "AttributeValue[1]"}


@mock_backend(occurrence_state_data)
def test_query_occurrences(blueprint_mock):

    cli = CliRunner()
    result = cli.invoke(
        query_occurrences,
        [
            "--query",
            "Occurrences[0].{ComponentId:Core.Component.Id}",
        ],
    )

    print(result.exc_info)
    assert result.exit_code == 0, result.output
    assert json.loads(result.output) == {"ComponentId": "ComponentId[1]"}



import os
import shutil
import tempfile

from cookiecutter.main import cookiecutter as cookiecutter_main
from hamlet.backend.common.exceptions import BackendException

try:
    from importlib.resources import files, as_file
except (ImportError, ModuleNotFoundError):
    from importlib_resources import files, as_file


def extract_package_to_temp(package, tmp_dir, root_dir, package_dir):
    for entry in files(package).joinpath(package_dir).iterdir():
        entry_path = os.path.commonprefix([root_dir, entry])
        entry_path = str(entry).replace(entry_path, "").lstrip("/")

        if entry.is_dir():
            os.makedirs(os.path.join(tmp_dir, entry_path))
            extract_package_to_temp(package, tmp_dir, root_dir, entry_path)
        else:
            tmp_file = os.path.join(tmp_dir, entry_path)
            shutil.copyfile(entry, tmp_file)


def replace_parameters_values(kwargs, replacers=None):
    for key, value in kwargs.items():
        for target, replacer in replacers:
            if value is target or value == target:
                kwargs[key] = replacer
                break


def cookiecutter(template_package, output_dir, **kwargs):

    with tempfile.TemporaryDirectory() as template_dir:

        if files(template_package).joinpath("cookiecutter.json").is_file():
            with as_file(
                files(template_package).joinpath("cookiecutter.json")
            ) as package_path:
                extract_package_to_temp(
                    template_package, template_dir, os.path.dirname(package_path), ""
                )
        else:
            raise BackendException(
                f"Provided template package does not contain cookiecutter.json config: {template_package}"
            )

        replace_parameters_values(kwargs, [[None, ""], [True, "yes"], [False, "no"]])
        cookiecutter_main(
            template_dir, no_input=True, output_dir=output_dir, extra_context=kwargs
        )



from hamlet.backend.common.exceptions import BackendException

import docker
import docker.errors


def run(SourceImage, DestinationImage, env={}):
    """
    push a docker image from the local docker host to a remote registry
    """

    client = docker.from_env()

    try:
        client.ping()
    except docker.errors.APIError as e:
        raise BackendException(str(e))

    try:
        source_image = client.images.get(SourceImage)

    except docker.errors.NotFound as e:
        raise BackendException(str(e))

    source_image.tag(DestinationImage)
    source_image.reload()
    client.images.push(DestinationImage)

    try:
        client.images.get_registry_data(DestinationImage)
    except docker.errors.NotFound as e:
        raise BackendException(str(e))

    return {
        "Properties": {
            "source_image": SourceImage,
            "destination_image": DestinationImage,
        }
    }



from hamlet.backend.automation_tasks.base import AutomationRunner
from hamlet.backend.automation import (
    properties_file,
    set_automation_context,
    manage_build_references,
    construct_tree,
)


class TransferImageAutomationRunner(AutomationRunner):
    def __init__(
        self,
        source_account,
        source_environment,
        deployment_unit,
        build_reference,
        image_format,
        engine,
        **kwargs
    ):
        _context_env = {
            "DEPLOYMENT_UNITS": deployment_unit,
            "IMAGE_FORMATS": image_format,
            "GIT_COMMIT": build_reference,
            "FROM_ACCOUNT": source_account,
            "FROM_ENVIRONMENT": source_environment,
            **kwargs,
        }

        super().__init__(engine=engine, **_context_env)

        self._script_list = [
            {
                "func": construct_tree.run,
                "args": {
                    "use_existing_tree": True,
                    "_is_cli": True,
                },
            },
            {
                "func": properties_file.get_automation_properties,
                "args": {**self._context_env},
            },
            {
                "func": properties_file.get_automation_properties,
                "args": {
                    **self._context_env,
                    "account": source_account,
                    "environment": source_environment,
                },
            },
            {
                "func": set_automation_context.run,
                "args": {"_is_cli": True, "release_mode": "promotion"},
            },
            {
                "func": manage_build_references.run,
                "args": {
                    "verify": "latest",
                    "code_commits": build_reference,
                    "deployment_units": deployment_unit,
                    "image_formats": image_format,
                    "_is_cli": True,
                },
            },
        ]



import os
import json
from unittest import mock
from click.testing import CliRunner
from hamlet.command.deploy.list import list_deployments


def template_backend_run_mock(data):
    def run(
        output_filename="unitlist-managementcontract.json",
        output_dir=None,
        *args,
        **kwargs
    ):
        os.makedirs(output_dir, exist_ok=True)
        unitlist_filename = os.path.join(output_dir, output_filename)
        with open(unitlist_filename, "wt+") as f:
            json.dump(data, f)

    return run


def mock_backend(unitlist=None):
    def decorator(func):
        @mock.patch("hamlet.backend.query.template")
        def wrapper(blueprint_mock, *args, **kwargs):

            blueprint_mock.run.side_effect = template_backend_run_mock(unitlist)

            return func(blueprint_mock, *args, **kwargs)

        return wrapper

    return decorator


@mock_backend(
    {
        "Stages": [
            {
                "Id": "StageId[1]",
                "Steps": [
                    {
                        "Id": "StepId[1]",
                        "Parameters": {
                            "DeploymentUnit": "DeploymentUnit[1]",
                            "DeploymentGroup": "DeploymentGroup[1]",
                            "DeploymentProvider": "DeploymentProvider[1]",
                            "Operations": ["Operation[1]"],
                            "CurrentState": "deployed",
                        },
                    },
                    {
                        "Id": "StepId[2]",
                        "Parameters": {
                            "DeploymentUnit": "DeploymentUnit[2]",
                            "DeploymentGroup": "DeploymentGroup[2]",
                            "DeploymentProvider": "DeploymentProvider[2]",
                            "Operations": ["Operation[2]"],
                            "CurrentState": "deployed",
                        },
                    },
                ],
            },
            {
                "Id": "StageId[2]",
                "Steps": [
                    {
                        "Id": "StepId[3]",
                        "Parameters": {
                            "DeploymentUnit": "DeploymentUnit[3]",
                            "DeploymentGroup": "DeploymentGroup[3]",
                            "DeploymentProvider": "DeploymentProvider[3]",
                            "Operations": ["Operation[3]"],
                            "CurrentState": "deployed",
                        },
                    },
                    {
                        "Id": "StepId[4]",
                        "Parameters": {
                            "DeploymentUnit": "DeploymentUnit[4]",
                            "DeploymentGroup": "DeploymentGroup[4]",
                            "DeploymentProvider": "DeploymentProvider[4]",
                            "Operations": ["Operation[4]"],
                            "CurrentState": "deployed",
                        },
                    },
                ],
            },
        ]
    }
)
def test_query_list_deployments(blueprint_mock):

    cli = CliRunner()
    result = cli.invoke(list_deployments, ["--output-format", "json"])
    print(result.output)
    assert result.exit_code == 0
    result = json.loads(result.output)
    assert len(result) == 4
    assert {
        "DeploymentUnit": "DeploymentUnit[1]",
        "DeploymentGroup": "DeploymentGroup[1]",
        "DeploymentProvider": "DeploymentProvider[1]",
        "Operations": ["Operation[1]"],
        "CurrentState": "deployed",
    } in result
    assert {
        "DeploymentUnit": "DeploymentUnit[2]",
        "DeploymentGroup": "DeploymentGroup[2]",
        "DeploymentProvider": "DeploymentProvider[2]",
        "Operations": ["Operation[2]"],
        "CurrentState": "deployed",
    } in result
    assert {
        "DeploymentUnit": "DeploymentUnit[3]",
        "DeploymentGroup": "DeploymentGroup[3]",
        "DeploymentProvider": "DeploymentProvider[3]",
        "Operations": ["Operation[3]"],
        "CurrentState": "deployed",
    } in result
    assert {
        "DeploymentUnit": "DeploymentUnit[4]",
        "DeploymentGroup": "DeploymentGroup[4]",
        "DeploymentProvider": "DeploymentProvider[4]",
        "Operations": ["Operation[4]"],
        "CurrentState": "deployed",
    } in result



from hamlet.backend.common.exceptions import BackendException
from base64 import b64decode

import boto3
import docker
import docker.errors
import subprocess


def run(
    RegistryId=None,
    Region=None,
    AWSAccessKeyId=None,
    AWSSecretAccessKey=None,
    AWSSessionToken=None,
    env={},
):
    """
    Perform a docker login for an AWS ECR Registry
    """

    docker_client = docker.from_env()

    try:
        docker_client.ping()
    except docker.errors.APIError as e:
        raise BackendException(str(e))

    session = boto3.Session(
        aws_access_key_id=AWSAccessKeyId,
        aws_secret_access_key=AWSSecretAccessKey,
        aws_session_token=AWSSessionToken,
        region_name=Region,
    )
    ecr = session.client("ecr")

    auth_token_args = {}
    if RegistryId:
        auth_token_args["registryIds"] = [RegistryId]

    auth_token_response = ecr.get_authorization_token(**auth_token_args)

    auth_token = auth_token_response["authorizationData"][0]

    ecr_password = (
        b64decode(auth_token["authorizationToken"])
        .replace(b"AWS:", b"")
        .decode("utf-8")
    )

    # See https://github.com/docker/docker-py/issues/1091
    # docker-py caches its login processes and they are only available for the lifetime
    # of the docker-py session
    # So we don't have to pass the docker client around we perform the docker login
    # directly using the docker command
    try:
        subprocess.run(
            [
                "docker",
                "login",
                "--username",
                "AWS",
                "--password-stdin",
                auth_token["proxyEndpoint"],
            ],
            input=ecr_password,
            check=True,
            text=True,
        )

    except subprocess.CalledProcessError as e:
        raise BackendException(str(e))

    return {"Properties": {"registry": auth_token["proxyEndpoint"]}}



import click

from hamlet.command import root as cli
from hamlet.command.common import exceptions, config

from hamlet.backend.automation_tasks import save_repos as save_repos_backend


@cli.group("cmdb", context_settings=dict(max_content_width=240))
def group():
    """
    Manage CMDBs
    """


@group.command(
    "commit-changes", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "--accounts/--no-accounts",
    default=False,
    help="Include account repostories in save",
)
@click.option(
    "--products/--no-products",
    default=True,
    help="Include product repositories in save",
)
@click.option(
    "--commit-type",
    help="The type of commit to assign these changes to",
    default="cd",
    type=click.Choice(
        [
            "cd",
            "feat",
            "chore",
            "fix",
            "refactor",
        ]
    ),
)
@click.option(
    "-d",
    "--commit-description",
    default="deploy_updates",
)
@click.option(
    "-m",
    "--commit-message",
    required=True,
    help="The commit message for these changes",
)
@click.option(
    "-b",
    "--branch",
    help="The repository reference to commit changes on",
)
@click.option(
    "--tag",
    help="The name of a tag to apply to the commit",
)
@click.option("--push/--no-push", default=True, help="Push committed changes")
@exceptions.backend_handler()
@config.pass_options
def commit_changes(
    options,
    accounts,
    products,
    commit_type,
    commit_description,
    commit_message,
    branch,
    tag,
    push,
    **kwargs,
):
    """
    Commit changes made to all CMDBs to their repositories
    """

    message = [
        f"cctype={commit_type}",
        f"ccdesc={commit_description}",
        f"msg={commit_message}",
    ]

    if accounts:
        message.append(f"account={options.account}")

    if products:
        message.append(f"product={options.product}")

    command_args = {
        "account_repos": accounts,
        "product_repos": products,
        "commit_message": ", ".join(message),
        "reference": branch,
        "defer_push": not push,
        "tag": tag,
    }

    task = save_repos_backend.SaveCMDBAutomationRunner(
        **options.opts, **kwargs, **command_args, engine=options.engine
    )
    task.run()



import click
import os
import json
from tabulate import tabulate

from hamlet.command.common.display import json_or_table_option, wrap_text
from hamlet.command import root as cli
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options
from hamlet.command.common.validate import validate_entrance_inputs
from hamlet.backend import query as query_backend
from hamlet.backend import contract as contract_backend


def query_runbookinfo_state(options, query, query_params=None, sub_query_text=None):

    query_args = {
        **options.opts,
        "generation_entrance": "runbookinfo",
        "output_filename": "runbookinfo-config.json",
    }
    query_result = query_backend.run(
        **query_args,
        cwd=os.getcwd(),
        engine=options.engine,
        query_text=query,
        query_params=query_params,
        sub_query_text=sub_query_text,
    )

    return query_result


LIST_RUNBOOKS_QUERY = (
    "RunBooks[]" ".{" "Name:Name," "Description:Description," "Engine:Engine" "}"
)

DESCRIBE_RUNBOOK_QUERY = "RunBooks[?Name=={name}] | [0]"


def list_runbooks_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["Name"]),
                wrap_text(row["Description"]),
                wrap_text(row["Engine"]),
            ]
        )
    return tabulate(
        tablerows,
        headers=[
            "Name",
            "Description",
            "Engine",
        ],
        tablefmt="github",
    )


@cli.group("task", context_settings=dict(max_content_width=240))
def group():
    """
    Runs tasks against a hamlet deployment
    """


@group.command(
    "list-runbooks", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-q", "--query", help="A JMESPath query to apply to the results")
@json_or_table_option(list_runbooks_table)
@exceptions.backend_handler()
@pass_options
def list_runbooks(options, query, **kwargs):
    """
    List available runbooks
    """
    return query_runbookinfo_state(
        options=options, query=LIST_RUNBOOKS_QUERY, sub_query_text=query
    )


@group.command(
    "describe-runbook", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-n", "--name", required=True, help="The name of the runbook to describe")
@click.option("-q", "--query", help="A JMESPath query to apply to the results")
@exceptions.backend_handler()
@pass_options
def describe_runbook(options, name, query=None, **kwargs):
    """
    Describe a specific runbook
    """

    result = query_runbookinfo_state(
        options=options,
        query=DESCRIBE_RUNBOOK_QUERY,
        query_params={"name": name},
        sub_query_text=query,
    )
    click.echo(json.dumps(result, indent=4))


@group.command(
    "run-runbook", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-n",
    "--name",
    required=True,
    help="The name of a runbook to execute",
)
@click.option(
    "--confirm/--no-confirm",
    default=False,
    help="Confirm before starting the runbook",
)
@click.option(
    "-s",
    "--silent",
    is_flag=True,
    default=False,
    help="Hide Runbook processing details",
)
@click.argument("inputs", nargs=-1, callback=validate_entrance_inputs)
@exceptions.backend_handler()
@pass_options
def run_runbook(options, name, confirm, silent, inputs, **kwargs):
    """
    Run a runbook
    Runbook inputs are provided as Key=Value inputs
    """

    runbook_description = query_runbookinfo_state(
        options=options,
        query=DESCRIBE_RUNBOOK_QUERY,
        query_params={"name": name},
    )

    if runbook_description is None:
        raise exceptions.CommandError("No runbooks found with the provided name")

    if not silent:
        click.echo("")
        click.secho(f"[*] {name}", bold=True, fg="green")
        if runbook_description["Description"] != "":
            click.secho(
                f"[*]   {runbook_description['Description']}", bold=True, fg="green"
            )
        click.echo("")

    if not silent or ((confirm and click.confirm("Start Runbook?")) or not confirm):

        query_args = {
            **options.opts,
            "generation_entrance": "runbook",
            "generation_entrance_parameter": (
                f"RunBook={name}",
                f"RunBookInputs={json.dumps(inputs)}",
            ),
            "output_filename": "runbook-contract.json",
        }
        contract = query_backend.run(
            **query_args, engine=options.engine, cwd=os.getcwd(), query=None
        )
        contract_backend.run(
            contract,
            silent,
            options.engine,
            {
                "HAMLET_HOME_DIR": options.opts.get("hamlet_home_dir", ""),
                "_engine": options.engine,
            },
        )



from typing import Any


def str_to_bool(value: Any) -> bool:
    """
    Return whether the provided string (or any value really) represents true. Otherwise false.
    Just like plugin server stringToBoolean.
    """
    if not value:
        return False
    return str(value).lower() in ("y", "yes", "t", "true", "on", "1")


def to_bool(value: Any, default) -> bool:
    """
    Converts string or None to bool
    """
    if isinstance(value, str):
        return str_to_bool(value)
    elif value is None:
        return default
    else:
        return value



import click
from hamlet.backend.run import task as run_task_backend
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options


@click.command(
    "task",
    short_help="Run an ECS task",
    context_settings=dict(max_content_width=240),
    deprecated=True,
)
@click.option(
    "-t",
    "--tier",
    required=True,
    help="name of the tier in the solution where the task is defined",
)
@click.option(
    "-i",
    "--component",
    required=True,
    help="name of the ecs component in the solution where the task is defined",
)
@click.option(
    "-j", "--component-instance", help="instance of the ecs cluster to run the task on"
)
@click.option(
    "-k", "--component-version", help="version of the ecs clsuter to run the task on"
)
@click.option(
    "-w",
    "--task",
    required=True,
    help="name of the task to be run",
)
@click.option("-x", "--instance", help="instance of the task to be run")
@click.option("-y", "--version", help="version of the task to be run")
@click.option(
    "-c",
    "--container-id",
    help="name of the container that environment details are applied to",
)
@click.option(
    "-e",
    "--env",
    "env_name",
    help="name of an environment variable to define for the task",
    multiple=True,
)
@click.option(
    "-v",
    "--value",
    help="value for the last environment value defined for the task",
    multiple=True,
)
@click.option(
    "-d",
    "--delay",
    help="interval between checking the progress of the task",
    type=click.INT,
    default=30,
    show_default=True,
)
@exceptions.backend_handler()
@pass_options
def task(options, **kwargs):
    """
    Run an ECS task

    \b
    NOTES:
    1. The ECS cluster is found using the provided tier and component combined with the product and segment
    2. ENV and VALUE should always appear in pairs
    """

    args = {**options.opts, **kwargs}

    run_task_backend.run(**args, engine=options.engine, _is_cli=True)



import json
import contextlib
import sys
import signal
import boto3
import errno
from subprocess import check_call


@contextlib.contextmanager
def ignore_user_entered_signals():
    """
    Ignores user entered signals to avoid process getting killed.
    """
    if sys.platform == "win32":
        signal_list = [signal.SIGINT]
    else:
        signal_list = [signal.SIGINT, signal.SIGQUIT, signal.SIGTSTP]
    actual_signals = []
    for user_signal in signal_list:
        actual_signals.append(signal.signal(user_signal, signal.SIG_IGN))
    try:
        yield
    finally:
        for sig, user_signal in enumerate(signal_list):
            signal.signal(user_signal, actual_signals[sig])


def invoke(
    session_id,
    stream_url,
    token_value,
    parameters,
    aws_access_key_id,
    aws_secret_access_key,
    aws_session_token,
    region_name,
    auth_profile_name="",
):
    """
    Invokes an SSM interactice session using session details
    that have been created in another process
    """

    session = boto3.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        region_name=region_name,
    )
    ssm = session.client("ssm")
    endpoint_url = ssm.meta.endpoint_url

    ssm_session = {
        "SessionId": session_id,
        "StreamUrl": stream_url,
        "TokenValue": token_value,
    }

    try:
        # ignore_user_entered_signals ignores these signals
        # because if signals which kills the process are not
        # captured would kill the foreground process but not the
        # background one. Capturing these would prevents process
        # from getting killed and these signals are input to plugin
        # and handling in there

        with ignore_user_entered_signals():
            # call executable with necessary input
            check_call(
                [
                    "session-manager-plugin",
                    json.dumps(ssm_session),
                    region_name,
                    "StartSession",
                    auth_profile_name,
                    json.dumps(parameters),
                    endpoint_url,
                ]
            )
        return 0

    except OSError as ex:
        if ex.errno == errno.ENOENT:

            # start-session api call returns response and starts the
            # session on ssm-agent and response is forwarded to
            # session-manager-plugin. If plugin is not present, terminate
            # is called so that service and ssm-agent terminates the
            # session to avoid zombie session active on ssm-agent for
            # default self terminate time
            ssm.terminate_session(SessionId=session_id)
            raise ValueError(
                (
                    "AWS Session manager plugin could not be found: "
                    "see https://docs.aws.amazon.com/systems-manager/"
                    "latest/userguide/session-manager-working-with-install-plugin.html"
                )
            )



# Project --------------------------------------------------------------

project = "Hamlet Python Executor"
copyright = "2021 Hamlet"
author = "Hamlet"

# General --------------------------------------------------------------

master_doc = "index"
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_click"
]
autodoc_typehints = "description"
intersphinx_mapping = {"python": ("https://docs.python.org/3/", None)}



from hamlet.backend.common import runner


def run(
    deployment_group=None,
    deployment_unit=None,
    build_logs=None,
    node_package_manager=None,
    binary_output_dir=None,
    log_level=None,
    log_format=None,
    root_dir=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    engine=None,
    _is_cli=False,
    **kwargs
):
    options = {
        "-l": build_logs,
        "-n": node_package_manager,
        "-o": binary_output_dir,
        "-u": deployment_unit,
        "-g": deployment_group,
    }
    env = {
        "GENERATION_LOG_LEVEL": log_level,
        "GENERATION_LOG_FORMAT": log_format,
        "ROOT_DIR": root_dir,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
    }
    runner.run(
        "runExpoAppPublish.sh",
        args=[],
        options=options,
        engine=engine,
        env=env,
        _is_cli=_is_cli,
    )



from hamlet.backend.common import runner


def run(
    entrance=None,
    config_ref=None,
    resource_group=None,
    request_ref=None,
    region=None,
    deployment_group=None,
    deployment_unit=None,
    deployment_unit_subset=None,
    deployment_mode=None,
    generation_provider=None,
    generation_framework=None,
    generation_input_source=None,
    output_dir=None,
    disable_output_cleanup=None,
    entrance_parameter=None,
    log_level=None,
    log_format=None,
    root_dir=None,
    tenant=None,
    district_type=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    engine=None,
    _is_cli=False,
    **kwargs,
):

    if isinstance(entrance_parameter, dict):
        entrance_parameter = tuple([f"{k}={v}" for k, v in entrance_parameter.items()])

    options = {
        "-e": entrance,
        "-c": config_ref,
        "-g": resource_group,
        "-l": deployment_group,
        "-q": request_ref,
        "-r": region,
        "-u": deployment_unit,
        "-z": deployment_unit_subset,
        "-d": deployment_mode,
        "-p": generation_provider,
        "-f": generation_framework,
        "-i": generation_input_source,
        "-o": output_dir,
        "-x": disable_output_cleanup,
        "-y": entrance_parameter,
    }
    env = {
        "GENERATION_LOG_LEVEL": log_level,
        "GENERATION_LOG_FORMAT": log_format,
        "ROOT_DIR": root_dir,
        "DISTRICT_TYPE": district_type,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
    }
    runner.run(
        "createTemplate.sh",
        args=[],
        options=options,
        env=env,
        engine=engine,
        _is_cli=_is_cli,
    )



from hamlet.backend.automation_tasks.base import AutomationRunner
from hamlet.backend.automation import (
    properties_file,
    set_automation_context,
    construct_tree,
    manage_images,
)


class UploadImageAutomationRunner(AutomationRunner):
    """
    Runs the automation tasks required to upload an image into a registry
    """

    def __init__(
        self,
        deployment_unit=None,
        build_reference=None,
        image_format=None,
        image_path=None,
        dockerfile=None,
        docker_context=None,
        docker_image=None,
        registry_scope=None,
        engine=None,
        **kwargs
    ):

        _context_env = {
            "DEPLOYMENT_UNITS": deployment_unit,
            "GIT_COMMIT": build_reference,
            "IMAGE_FORMATS": image_format,
            "REGISTRY_SCOPE": registry_scope,
            **kwargs,
        }

        super().__init__(engine=engine, **_context_env)

        self._script_list = [
            {
                "func": construct_tree.run,
                "args": {
                    "use_existing_tree": True,
                    "_is_cli": True,
                },
            },
            {
                "func": properties_file.get_automation_properties,
                "args": {**self._context_env},
            },
            {"func": set_automation_context.run, "args": {"_is_cli": True}},
            {
                "func": manage_images.run,
                "args": {
                    "registry_scope": registry_scope,
                    "dockerfile": dockerfile,
                    "docker_context": docker_context,
                    "image_formats": image_format,
                    "code_commit": build_reference,
                    "image_paths": image_path,
                    "docker_image": docker_image,
                    "deployment_unit": deployment_unit,
                    "_is_cli": True,
                },
            },
        ]



import shutil


def run(currentFileName, newFileName, env={}):
    """
    rename a file by moving it between two locations
    """

    shutil.move(currentFileName, newFileName)

    return {"Properties": {}}



import typing
import os
import json

from hamlet.backend.engine.engine_loader import EngineLoader

from hamlet.backend.engine.engine import Engine, ShimEngine, InstalledEngine
from hamlet.backend.engine.engine_part import (
    CoreEnginePart,
    AWSEnginePluginPart,
    AzureEnginePluginPart,
    CMDBEnginePluginPart,
    BashExecutorEnginePart,
    WrapperEnginePart,
)
from hamlet.backend.engine.engine_source import ShimPathEngineSource

from hamlet.backend.engine.common import (
    ENGINE_STATE_FILE_NAME,
)


class ShimEngineLoader(EngineLoader):
    """
    Seeds an engine used to create a known engine directory that can be used to establish
    a standard location for hamlet scripts called outside of the cli
    """

    def load(self) -> typing.Iterable[Engine]:
        engine_source = [
            ShimPathEngineSource(name="shim", description="shim based root dir source")
        ]

        engine_parts = [
            CoreEnginePart(source_path="engine-core", source_name="shim"),
            BashExecutorEnginePart(source_path="executor-bash", source_name="shim"),
            AWSEnginePluginPart(source_path="engine-plugin-aws", source_name="shim"),
            AzureEnginePluginPart(
                source_path="engine-plugin-azure", source_name="shim"
            ),
            CMDBEnginePluginPart(source_path="engine-plugin-cmdb", source_name="shim"),
            WrapperEnginePart(source_path="engine-wrapper", source_name="shim"),
        ]

        engine = ShimEngine(
            name="shim",
            description="The engine used to provide shim based access to a fixed diretory",
        )
        engine.parts = engine_parts
        engine.sources = engine_source

        yield engine


class InstalledEngineLoader(EngineLoader):
    """
    Discovers any existing engines under a directory and seeds them as basic
    engine implementations
    """

    def __init__(self, engine_dir):
        super().__init__()
        self.engine_dir = engine_dir

    def load(self) -> typing.Iterable[Engine]:

        engine_states = []
        if os.path.isdir(self.engine_dir):
            with os.scandir(self.engine_dir) as engines:
                for engine in engines:
                    if engine.is_dir():
                        with os.scandir(engine) as sources:
                            for source in sources:
                                if (
                                    source.is_file()
                                    and source.name == ENGINE_STATE_FILE_NAME
                                ):
                                    with open(source, "r") as f:
                                        engine_states.append(json.load(f))

        for engine_state in engine_states:
            engine = InstalledEngine(
                name=engine_state["name"],
                description=engine_state["description"],
                digest=engine_state["install"]["digest"],
                state_version=engine_state.get("version", "0.0.0"),
            )
            yield engine



import click


def validate_entrance_inputs(ctx, param, value):
    """
    Custom validation for entrance inputs which are provided as Key=Value
    strings
    """
    try:
        value = dict([arg.split("=", 1) for arg in value])
    except ValueError:
        raise click.BadParameter("input arguments must be in 'Key=Value' format")

    return value



import os
import shutil
import pytest
from unittest import mock
from hamlet.loggers import logging, root
from hamlet.backend.engine import EngineStore
from hamlet.command.common.config import Options


root.setLevel(logging.DEBUG)


class __CMDBContext:

    if os.getenv("TEST_ROOT_DIR", None) is not None:
        ROOT_DIR = os.getenv("TEST_ROOT_DIR")
    else:
        ROOT_DIR = os.getcwd() + "/.cmdb"
        os.makedirs(ROOT_DIR, exist_ok=True)

    def create_cmdb_filename_compositor(self):
        root = self.dir

        def cmdb_filename_compositor(*args, sep="-", ext=""):
            name = sep.join([*args])
            filename = os.path.join(root, name)
            if ext:
                filename += ".%s" % ext
            return os.path.normpath(filename)

        return cmdb_filename_compositor

    def __init__(self, *args):
        self.dir = os.path.join(self.ROOT_DIR, *args)
        self.old_cwd = os.getcwd()

    def __enter__(self):
        assert os.path.exists(self.dir), "Path %s does not exit" % self.dir
        assert os.path.isdir(self.dir), "Path %s is not a dir" % self.dir
        os.chdir(self.dir)
        return self.create_cmdb_filename_compositor()

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.old_cwd)


def __clear_dir(dir):
    assert os.path.isdir(dir)
    for entry in os.listdir(dir):
        entry = os.path.abspath(os.path.join(dir, entry))
        if os.path.isdir(entry):
            shutil.rmtree(entry)
        else:
            os.remove(entry)


def __clear_cmdb(*args):
    with __CMDBContext() as path:
        if not args:
            __clear_dir(path())
        else:
            __clear_dir(os.path.join(*args))


@pytest.fixture(scope="session")
def cmdb():
    return __CMDBContext


@pytest.fixture(scope="session")
def clear_cmdb():
    return __clear_cmdb


class __EngineContext:
    def __init__(self) -> None:
        if os.getenv("TEST_ENGINE_STORE_DIR", None) is not None:
            ENGINE_STORE_DIR = os.getenv("TEST_ENGINE_STORE_DIR")
        else:
            ENGINE_STORE_DIR = os.getcwd() + "/.engine_store"
            os.makedirs(ENGINE_STORE_DIR, exist_ok=True)

        self._engine_store = EngineStore(
            store_dir=ENGINE_STORE_DIR, config_search_paths=ENGINE_STORE_DIR
        )
        self._engine_store.load_engines(locations=["local", "remote"], refresh=True)
        self._engine_store.get_engine("unicycle", ["remote"]).install()
        self._engine_store.load_engines(locations=["installed"], refresh=True)

        self._engine = self._engine_store.get_engine("unicycle", ["installed"])

    @property
    def engine(self):
        return self._engine

    @property
    def engine_store(self):
        return self._engine_store


@pytest.fixture(scope="session")
def engine():
    return __EngineContext()


def mock_engine():
    engine = mock.MagicMock()
    engine.name = mock.PropertyMock().return_value = "Name[1]"
    engine.location = mock.PropertyMock().return_value = "local"
    engine.description = mock.PropertyMock().return_value = "Description[1]"
    engine.short_digest = mock.PropertyMock().return_value = "Digest[1]"
    engine.digest = mock.PropertyMock().return_value = "Digest[1]"
    return engine


def mock_engine_store():
    engine_store = mock.MagicMock()
    engine_store.get_engines.return_value = [mock_engine()]
    return engine_store


class __Options(Options):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._engine = mock_engine()
        self._engine_store = mock_engine_store()


@pytest.fixture(scope="session")
def options():
    return __Options()


def __clear_engine(*args):
    if os.path.exists(__EngineContext.ENGINE_STORE_DIR) and os.path.isdir(
        __EngineContext.ENGINE_STORE_DIR
    ):
        shutil.rmtree(__EngineContext.ENGINE_STORE_DIR)


@pytest.fixture(scope="session")
def clear_engine():
    return __clear_engine



import os
import json
import tempfile
import pytest
from hamlet.backend.common.fsutils import Search, ContextSearch, Directory, File


def test_search():
    with tempfile.TemporaryDirectory() as temp_dir:
        path = "/test/path/split"
        assert Search.split_path(path) == ["/", "test", "path", "split"]
        path = "test/path/split"
        assert Search.split_path(path) == ["test", "path", "split"]

        test_directory_fullpath = os.path.join(temp_dir, "test/directory")
        test_directory_1_fullpath = os.path.join(temp_dir, "test/directory/1")
        test_directory_2_fullpath = os.path.join(temp_dir, "test/directory/2")
        test_filename_fullpath = os.path.join(test_directory_fullpath, "node-0")
        test_filename_1_fullpath = os.path.join(test_directory_1_fullpath, "node-1")
        test_filename_2_fullpath = os.path.join(test_directory_2_fullpath, "node-2")
        os.makedirs(test_directory_1_fullpath)
        os.makedirs(test_directory_2_fullpath)
        open(test_filename_fullpath, "w").close()
        open(test_filename_1_fullpath, "w").close()
        open(test_filename_2_fullpath, "w").close()

        assert os.path.exists(test_filename_fullpath)
        assert os.path.exists(test_filename_1_fullpath)
        assert os.path.exists(test_filename_2_fullpath)

        with pytest.raises(ValueError):
            ContextSearch(os.path.relpath(temp_dir, test_directory_fullpath))

        with pytest.raises(ValueError):
            ContextSearch(test_filename_fullpath)

        with pytest.raises(ValueError):
            ContextSearch("/directory/doesnt/exist")

        assert (
            Search.exists(test_directory_fullpath, "node-0") == test_filename_fullpath
        )
        assert (
            ContextSearch(test_directory_fullpath).exists("node-0")
            == test_filename_fullpath
        )
        assert (
            Search.exists(test_directory_1_fullpath, "node-1")
            == test_filename_1_fullpath
        )
        assert (
            ContextSearch(test_directory_1_fullpath).exists("node-1")
            == test_filename_1_fullpath
        )
        assert (
            Search.exists(test_directory_2_fullpath, "node-2")
            == test_filename_2_fullpath
        )
        assert (
            ContextSearch(test_directory_2_fullpath).exists("node-2")
            == test_filename_2_fullpath
        )

        assert (
            Search.exists(test_directory_1_fullpath, "node-0", up=1)
            == test_filename_fullpath
        )
        assert (
            ContextSearch(test_directory_1_fullpath).exists("node-0", up=1)
            == test_filename_fullpath
        )
        assert (
            Search.exists(test_directory_2_fullpath, "node-0", up=1)
            == test_filename_fullpath
        )
        assert (
            ContextSearch(test_directory_2_fullpath).exists("node-0", up=1)
            == test_filename_fullpath
        )

        assert (
            Search.isfile(test_directory_fullpath, "node-0") == test_filename_fullpath
        )
        assert (
            ContextSearch(test_directory_fullpath).isfile("node-0")
            == test_filename_fullpath
        )

        assert (
            Search.isfile(test_directory_1_fullpath, "node-0", up=1)
            == test_filename_fullpath
        )
        assert (
            ContextSearch(test_directory_1_fullpath).isfile("node-0", up=1)
            == test_filename_fullpath
        )

        assert (
            Search.isdir(
                test_directory_fullpath, os.path.basename(test_directory_1_fullpath)
            )
            == test_directory_1_fullpath
        )
        assert (
            ContextSearch(test_directory_fullpath).isdir(
                os.path.basename(test_directory_1_fullpath)
            )
            == test_directory_1_fullpath
        )
        assert (
            Search.isdir(
                test_directory_1_fullpath,
                os.path.basename(test_directory_fullpath),
                up=2,
            )
            == test_directory_fullpath
        )
        assert (
            ContextSearch(test_directory_1_fullpath).isdir(
                os.path.basename(test_directory_fullpath), up=2
            )
            == test_directory_fullpath
        )

        assert (
            Search.upwards(test_directory_1_fullpath, "node-0")
            == test_filename_fullpath
        )
        assert (
            ContextSearch(test_directory_2_fullpath).upwards("node-0")
            == test_filename_fullpath
        )

        assert Search.downwards(temp_dir, "node-1") == [test_filename_1_fullpath]
        assert ContextSearch(temp_dir).downwards("node-2") == [test_filename_2_fullpath]

        assert Search.basename(test_directory_fullpath) == os.path.basename(
            test_directory_fullpath
        )
        assert ContextSearch(test_directory_fullpath).basename() == os.path.basename(
            test_directory_fullpath
        )

        assert Search.basename(test_directory_1_fullpath, up=1) == os.path.basename(
            test_directory_fullpath
        )
        assert ContextSearch(test_directory_2_fullpath).basename(
            up=1
        ) == os.path.basename(test_directory_fullpath)

        assert Search.parent(test_directory_1_fullpath, up=1) == test_directory_fullpath
        assert (
            ContextSearch(test_directory_2_fullpath).parent(up=1)
            == test_directory_fullpath
        )

        assert Search.parent(test_directory_1_fullpath, up=3) == temp_dir
        assert ContextSearch(test_directory_2_fullpath).parent(up=3) == temp_dir

        assert (
            Search.cut(
                "/p/re/fix/path/to/something/suf/fix",
                prefix="/p/re/fix",
                suffix="suf/fix",
            )
            == "path/to/something"
        )
        assert (
            Search.cut("/p/re/fix/path/to/something/suf/fix", prefix="/p/re/fix")
            == "path/to/something/suf/fix"
        )
        assert (
            Search.cut("/p/re/fix/path/to/something/suf/fix", suffix="suf/fix")
            == "/p/re/fix/path/to/something"
        )


def test_fs_interface():
    with tempfile.TemporaryDirectory() as temp_dir:
        test_directory_relpath = os.path.join("test", "directory", "path")
        test_directory_fullpath = os.path.join(temp_dir, test_directory_relpath)
        test_json_file_fullpath = os.path.join(test_directory_fullpath, "test.json")
        test_text_file_fullpath = os.path.join(test_directory_fullpath, "test.txt")
        os.makedirs(test_directory_fullpath)
        with open(test_json_file_fullpath, "wt+") as f:
            data = {"data": {"key": "value"}}
            json.dump(data, f)
        with open(test_text_file_fullpath, "wt+") as f:
            f.write("Hello world!")

        directory = Directory(temp_dir)
        assert (
            directory["test"]["directory"]["path"]["test.json"]["data"]["key"]
            == "value"
        )
        with pytest.raises(KeyError):
            assert directory["test"]["directory"]["path"]["nofile.json"]
        with pytest.raises(TypeError) as einfo:
            directory["test"]["directory"]["path"]["test.txt"]["data"]
        assert str(einfo.value) == "Unstructured data"

    with tempfile.TemporaryDirectory() as temp_dir:
        for i in range(10):
            open(os.path.join(temp_dir, f"{i}-file"), "w").close()
            os.mkdir(os.path.join(temp_dir, f"{i}-directory"))
        directory = Directory(temp_dir)
        c = 0
        for item in directory:
            c += 1
            if item.path.endswith("-file"):
                assert isinstance(item, File)
            elif item.path.endswith("-directory"):
                assert isinstance(item, Directory)
        assert c == 20

    with tempfile.TemporaryDirectory() as temp_dir:
        test_json_file_fullpath = os.path.join(temp_dir, "test.json")
        test_text_file_fullpath = os.path.join(temp_dir, "test.txt")
        with open(test_json_file_fullpath, "wt+") as f:
            data = {"data": {"key": "value"}}
            json.dump(data, f)
        with open(test_text_file_fullpath, "wt+") as f:
            f.write("Hello world!")

        file_interface = File(test_json_file_fullpath)
        assert file_interface["data"]["key"] == "value"
        file_interface["data"]["key"] = "modified"
        file_interface.write()
        file_interface = File(test_json_file_fullpath)
        assert file_interface["data"]["key"] == "modified"
        with open(test_json_file_fullpath, "wt+") as f:
            data = {"message": "changed"}
            json.dump(data, f)
        file_interface.reload()
        assert file_interface["message"] == "changed"

        file_interface = File(test_text_file_fullpath)
        file_interface.load()
        assert file_interface.data == "Hello world!"
        file_interface.data = "Goodbye world!"
        file_interface.write()
        file_interface = File(test_text_file_fullpath)
        file_interface.load()
        assert file_interface.data == "Goodbye world!"



import boto3
from hamlet.backend.common import aws_ssm_session


def build_ssm_request_paramaters(task_details, container_name):
    cluster_name = task_details["clusterArn"].split("/")[-1]
    task_id = task_details["taskArn"].split("/")[-1]

    container_details = list(
        filter(
            lambda container: container["name"] == container_name,
            task_details["containers"],
        )
    )[0]

    ssm_request_params = {
        "Target": f"ecs:{cluster_name}_{task_id}_{container_details['runtimeId']}"
    }
    return ssm_request_params


def run(
    ClusterArn=None,
    TaskArn=None,
    ContainerName=None,
    Command=None,
    Region=None,
    AWSAccessKeyId=None,
    AWSSecretAccessKey=None,
    AWSSessionToken=None,
    env={},
):
    """
    Run an interactive command on aws ecs task
    Uses ECS SSM integration to provide interactive console through the
    Session Manager plugin for aws cli
    """
    session = boto3.Session(
        aws_access_key_id=AWSAccessKeyId,
        aws_secret_access_key=AWSSecretAccessKey,
        aws_session_token=AWSSessionToken,
        region_name=Region,
    )

    ecs = session.client("ecs")

    execute_params = {
        "cluster": ClusterArn,
        "task": TaskArn,
        "command": Command,
        "interactive": True,
    }

    if ContainerName is not None:
        execute_params["container"] = ContainerName

    command = ecs.execute_command(**execute_params)

    task_details = ecs.describe_tasks(cluster=ClusterArn, tasks=[TaskArn])
    task_details = list(
        filter(lambda task: task["taskArn"] == TaskArn, task_details["tasks"])
    )[0]

    if ContainerName is None:
        ContainerName = task_details["containers"][0]["name"]

    aws_ssm_session.invoke(
        session_id=command["session"]["sessionId"],
        stream_url=command["session"]["streamUrl"],
        token_value=command["session"]["tokenValue"],
        parameters=build_ssm_request_paramaters(task_details, ContainerName),
        aws_access_key_id=AWSAccessKeyId,
        aws_secret_access_key=AWSSecretAccessKey,
        aws_session_token=AWSSessionToken,
        region_name=Region,
    )

    return {"Properties": {}}



import io
from fabric import Connection
from paramiko import RSAKey
from invoke.exceptions import UnexpectedExit
from socket import timeout as socket_timeout

from hamlet.backend.contract.tasks.exceptions import TaskFailureException


def get_connect_args(password, ssh_key):
    connect_kwargs = {"allow_agent": False, "look_for_keys": False}

    if password is not None:
        connect_kwargs["password"] = password

    if ssh_key is not None:

        if isinstance(ssh_key, str):
            private_key = RSAKey.from_private_key(io.StringIO(ssh_key))
        else:
            private_key = RSAKey.from_private_key(ssh_key)

        connect_kwargs["pkey"] = private_key

    return connect_kwargs


def run(
    Host,
    Port,
    Username,
    Password=None,
    SSHKey=None,
    Command=None,
    BastionHost=None,
    BastionPort=None,
    BastionUsername=None,
    BastionPassword=None,
    BastionSSHKey=None,
    env={},
):
    """
    Run a command on a remote ssh host with an interactive shell
    Command can also be proxied through a bastion gateway
    """

    print(f"\n---- SSH Session | host: {Host} | user: {Username} ----\n")

    gateway_connection = None

    if BastionHost is not None:
        gateway_connection = Connection(
            host=BastionHost,
            port=BastionPort,
            user=BastionUsername,
            connect_kwargs=get_connect_args(BastionPassword, BastionSSHKey),
            connect_timeout=10,
        )

    with Connection(
        host=Host,
        user=Username,
        port=Port,
        connect_kwargs=get_connect_args(Password, SSHKey),
        connect_timeout=10,
        gateway=gateway_connection,
    ) as ssh_con:
        try:
            ssh_con.run(Command, pty=True)
        except socket_timeout as e:
            raise TaskFailureException(str(e))
        except UnexpectedExit:
            pass

    print("\n---- SSH Session Complete ----\n")

    return {}



import click
import os
import json
from tabulate import tabulate

from hamlet.command import root as cli
from hamlet.command.common import config, exceptions
from hamlet.command.common.display import json_or_table_option, wrap_text

from hamlet.backend.engine.engine_code_source import EngineCodeSourceBuildData
from hamlet.backend.engine.exceptions import (
    HamletEngineInvalidVersion,
    EngineStoreMissingEngineException,
)


def engine_locations_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["name"]),
                wrap_text(row["description"]),
            ]
        )
    return tabulate(
        tablerows,
        headers=["Name", "Description"],
        tablefmt="github",
    )


def engines_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["name"]),
                wrap_text(row["location"]),
                wrap_text(row["description"]),
                wrap_text(row["short_digest"]),
            ]
        )
    return tabulate(
        tablerows,
        headers=["Name", "Location", "Description", "Short Digest"],
        tablefmt="github",
    )


@cli.group("engine", context_settings=dict(max_content_width=240))
def group():
    """
    Manage the engine used by the executor
    """
    pass


@group.command(
    "list-engine-locations", short_help="", context_settings=dict(max_content_width=240)
)
@json_or_table_option(engine_locations_table)
@exceptions.backend_handler()
@config.pass_options
def list_engine_locations(options):
    """
    Lists the engine locations
    """
    data = []
    for k, location in options.engine_store.engine_locations.items():
        data.append(
            {
                "name": k,
                "description": location["description"],
            }
        )

    return data


@group.command(
    "list-engines", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-l",
    "--location",
    multiple=True,
    help="The location of the engine to look for the engine",
    default=["installed", "local", "remote"],
)
@json_or_table_option(engines_table)
@exceptions.backend_handler()
@config.pass_options
def list_engines(options, location):
    """
    Lists the engines available
    """
    options.engine_store.load_engines(locations=location, refresh=True)
    data = []
    for engine_instance in options.engine_store.get_engines(locations=location):
        data.append(
            {
                "name": engine_instance.name,
                "location": engine_instance.location,
                "description": engine_instance.description,
                "short_digest": engine_instance.short_digest,
                "digest": engine_instance.digest,
            }
        )

    return data


@group.command(
    "describe-engine", short_help="", context_settings=dict(max_content_width=240)
)
@click.option("-l", "--location", help="The location of the engine to describe")
@click.argument("name", required=False, type=click.STRING)
@exceptions.backend_handler()
@config.pass_options
def describe_engine(options, name, location):
    """
    Provides a detailed description of an engine
    """

    location = None if not location else [location]

    if name:
        options.engine_store.load_engines(locations=location, refresh=True)
        engines = [
            x
            for x in options.engine_store.get_engines(locations=location)
            if x.name == name
        ]
        if len(engines) > 1:

            extra_engines = "\n".join(
                [f" - name: {x.name} - location: {x.location}" for x in engines]
            )

            raise click.exceptions.UsageError(
                (
                    f"Multiple Engines found for the provided name: {name}\n"
                    f"{extra_engines}\n"
                    "run the command with the --location option to pick an engine"
                )
            )
        else:
            engine_instance = engines[0]

    else:
        engine_instance = options.engine

    engine_details = {
        "engine": {
            "name": engine_instance.name,
            "description": engine_instance.description,
            "location": engine_instance.location,
            "engine_dir": engine_instance.engine_dir,
            "digest": engine_instance.digest,
        },
        "environment": engine_instance.environment,
        "install_state": engine_instance.install_state,
    }

    sources = []
    for source in engine_instance.sources:

        try:
            source_digest = source.digest
        except BaseException as e:
            click.secho(
                f"[!] Source check failed {engine_instance.name} - {source.name}",
                fg="red",
                err=True,
            )
            click.secho(f"[!]  {e}", fg="red", err=True)

            source_digest = None
            pass

        sources.append(
            {
                "name": source.name,
                "description": source.description,
                "latest_digest": source_digest,
            }
        )

    parts = []
    for part in engine_instance.parts:
        parts.append(
            {
                "type": part.type,
                "description": part.description,
                "source_path": part.source_path,
                "source_name": part.source_name,
            }
        )

    engine_details["sources"] = sources
    engine_details["parts"] = parts

    click.echo(json.dumps(engine_details, indent=2))


@group.command(
    "clean-engines", short_help="", context_settings=dict(max_content_width=240)
)
@click.argument("name", type=click.STRING, nargs=-1)
@exceptions.backend_handler()
@config.pass_options
def clean_engines(options, name):
    """
    Clean local engine store
    """
    if name:
        for x in name:
            click.echo(f"[*] cleaning {x} from {options.engine_store.store_dir}")
            options.engine_store.clean_engine(x)

            if options.engine_store.default_engine == name:
                options.engine_store.default_engine = None

    else:
        click.echo(f"[*] cleaning all engines from {options.engine_store.store_dir}")
        options.engine_store.clean_engines()
        options.engine_store.default_engine = None


@group.command(
    "install-engine", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-u",
    "--update",
    is_flag=True,
    help="Update the engine if it is installed",
)
@click.option(
    "-l",
    "--location",
    multiple=True,
    help="The location of the engine to install",
    default=["local", "remote", "hidden"],
)
@click.argument("name", required=False, type=click.STRING)
@exceptions.backend_handler()
@config.pass_options
def install_engine(options, name, location, update):
    """
    Install an engine
    """

    options.engine_store.load_engines(locations=location, refresh=True)

    if name is None and options.engine:
        name = options.engine.name

    if name is None:
        raise click.exceptions.BadParameter(
            "Engine name missing - provide name with HAMLET_ENGINE or as an argument"
        )

    try:
        engine_instance = options.engine_store.get_engine(name, locations=location)

    except HamletEngineInvalidVersion as e:
        click.secho(
            (
                "[!] The state of this engine is not compatible with the cli\n"
                f"[!] Remove the engine using\n"
                f"     hamlet engine clean-engine {name}\n"
                "[!] then install the engine again"
            ),
            err=True,
            fg="red",
        )
        raise e

    try:
        installed_engine = options.engine_store.get_engine(
            name, locations=["installed"]
        )
    except EngineStoreMissingEngineException:
        installed_engine = False
        pass

    if installed_engine and update:
        engine_instance.install()
        click.echo(f"[*] updating engine | {name} | digest: {engine_instance.digest}")
    elif not installed_engine:
        engine_instance.install()
        click.echo(f"[*] installing engine | {name} | digest: {engine_instance.digest}")
    else:
        click.echo(
            f"[*] skipping engine update | {name} | digest: {engine_instance.digest}"
        )


@group.command(
    "set-engine", short_help="", context_settings=dict(max_content_width=240)
)
@click.argument("name", required=True, type=click.STRING)
@exceptions.backend_handler()
@config.pass_options
def set_engine(options, name):
    """
    Sets the default engine
    """
    engine = options.engine_store.get_engine(name, locations=["installed"])
    options.engine_store.default_engine = engine.name

    click.echo(f"[*] default engine set to {engine.name}")


@group.command(
    "get-engine", short_help="", context_settings=dict(max_content_width=240)
)
@exceptions.backend_handler()
@config.pass_options
def get_engine(options):
    """
    Gets the current global engine
    """
    name = (
        options.engine.name if options.engine else options.engine_store.default_engine
    )
    click.echo(options.engine_store.get_engine(name, locations=["installed"]).name)


@group.command("env", short_help="", context_settings=dict(max_content_width=240))
@click.argument(
    "environment_variable",
    required=False,
    type=click.STRING,
)
@exceptions.backend_handler()
@config.pass_options
def env(options, environment_variable):
    """
    Get the environment variables for the current engine
    """

    if environment_variable is None:
        click.echo("# run eval $(hamlet engine env) to set variables")
        for k, v in options.engine.environment.items():
            click.echo(f'export {k}="{v}"')

    else:
        try:
            click.echo(options.engine.environment[environment_variable])
        except KeyError:
            click.echo("")


@group.command(
    "add-engine-source-build",
    short_help="",
    context_settings=dict(max_content_width=240),
)
@click.option(
    "-p",
    "--path",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
    default=".",
    help="The path to generate build details for",
)
@exceptions.backend_handler()
def add_engine_source_build(path):
    """
    Generates build metadata for engine sources
    """

    build_details = EngineCodeSourceBuildData(path=path)

    hamlet_meta_dir = os.path.join(path, ".hamlet")
    hamlet_build_state_file = os.path.join(hamlet_meta_dir, "engine_source.json")
    if not os.path.isdir(hamlet_meta_dir):
        os.makedirs(hamlet_meta_dir)

    with open(hamlet_build_state_file, "w") as file:
        json.dump(build_details.details, file, indent=2)



import os
from .exceptions import BackendException
from .fsutils import ContextSearch, Search, File


# This is a legacy module that soon must be replaced by a new class based on results of direct shell scrips rewrite


class ContextError(BackendException):
    def __init__(self, msg="Context error"):
        super().__init__(msg)


class NoLevelFileError(ContextError):
    def __init__(self, msg="Level file not found"):
        super().__init__(msg)


class NoRootFileError(ContextError):
    def __init__(self, msg="Root file not found"):
        super().__init__(msg)


class MultipleTenantsFoundError(ContextError):
    def __init__(self, msg="Multiple tenants found"):
        super().__init__(msg)


class NoTenantFoundError(ContextError):
    def __init__(self, msg="Tenant not found"):
        super().__init__(msg)


class NoAccountsFoundError(ContextError):
    def __init__(self, msg="No accounts found"):
        super().__init__(msg)


class MultipleAccountsFoundError(ContextError):
    def __init__(self, msg="Multiple accounts found"):
        super().__init__(msg)


class SpecifiedAccountNotFoundError(ContextError):
    def __init__(self, msg="Specified account not found"):
        super().__init__(msg)


class Context:
    # Enables tenant->account lookup at the init stage
    ACCOUNT_REQUIRED = False
    # Enables tenant lookup at the init stage
    TENANT_REQUIRED = False
    # Context level name
    level_name = None
    # Context level_file
    level_file = None
    # Context level_file directory
    level_file_directory = None

    # Directory used for initializing context instance
    @property
    def directory(self):
        return self.__directory

    # Backend CWD. It can be changed unlike context initialization directory.
    @property
    def cwd(self):
        return self.__cwd

    @cwd.setter
    def cwd(self, value):
        self.__cwd = value

    # Root file directory
    @property
    def root_dir(self):
        return self.__root_dir

    @root_dir.setter
    def root_dir(self, value):
        self.__root_dir = value

    @property
    def account_dir(self):
        return self._account_dir

    @property
    def tenant_dir(self):
        return self._tenant_dir

    @property
    def product_dir(self):
        return self._product_dir

    # Context search tied to current directory
    @property
    def search(self):
        return self.__search

    def setup(self):
        pass

    def __init__(
        self,
        directory: str,
        root_dir: str = None,
        config: dict = None,
    ):
        config = config or {}
        self.config = config
        self.props = {}
        self.__search = ContextSearch(directory)
        self.__directory = directory
        self.__cwd = directory
        # semi private variables
        self._product_dir = None
        self._account_dir = None
        self._tenant_dir = None
        if root_dir is not None:
            self.__root_dir = root_dir
        else:
            self.__try_to_find_root()

        if self.ACCOUNT_REQUIRED or self.TENANT_REQUIRED:
            self.__try_to_set_tenant()
            if self.ACCOUNT_REQUIRED:
                self.__try_to_set_account()
        if self.level_name:
            self.__try_to_find_level_file()
        self.setup()

    def __try_to_find_root(self):
        if RootLevel.level_file_directory:
            found = self.search.downwards(
                RootLevel.level_file_directory, RootLevel.level_file
            )
        found = self.search.upwards(RootLevel.level_file)
        if not found:
            raise NoRootFileError(f"Can't find {RootLevel.level_file} file.")
        self.__root_dir = Search.parent(found, up=1)

    def __try_to_set_account(self):
        # Trying to find account directory in tenant directory(props['tenant'])
        # If there is a single account and suggested account is not set(config['account'])
        # found account will be used by default
        # If suggested account is set(config['account']), try to find account with name == config['account']
        # If no account with that name found raise SpecifiedAccountNotFoundError.
        accounts_directory = Search.parent(self.tenant_dir, up=1)
        account_suffix = os.path.join("config", AccountLevel.level_file)

        # Each account filepath has common prefix and suffix
        def name_from_path(path):
            return Search.cut(path, prefix=accounts_directory, suffix=account_suffix)

        found = Search.downwards(accounts_directory, AccountLevel.level_file)
        account_name = self.config.get("account")
        account_path = None
        if not found:
            raise NoAccountsFoundError(f"Can't find accounts in {self.tenant_dir}")
        elif len(found) == 1:
            account_path = found[0]
            if account_name and account_name != name_from_path(account_path):
                raise SpecifiedAccountNotFoundError(
                    f"Can't find account {account_name} in {accounts_directory}"
                )
        else:
            if not account_name:
                raise MultipleAccountsFoundError(
                    f"Multiple account found in {accounts_directory}. Specify account name"
                )
            for account_path in found:
                if account_name == name_from_path(account_path):
                    break
            else:
                raise SpecifiedAccountNotFoundError(
                    f"Can't find account {account_name} in {accounts_directory}"
                )
        self._account_dir = Search.parent(account_path, up=2)

    def __try_to_set_tenant(self):
        # Searching for tenant directory. Directory should contain TenantLevel.level_file
        # Multiple tenants not supported at the moment
        found = Search.downwards(self.root_dir, TenantLevel.level_file)
        if not found:
            raise NoTenantFoundError()
        elif len(found) > 1:
            raise MultipleTenantsFoundError()
        else:
            self._tenant_dir = Search.parent(found[0], up=1)

    def __try_to_find_level_file(self):
        # Try to find level file, path to which is set by directory + level_file_directory + level_file
        # Used to deremine the correctness of suggested level.
        # If level file not found, raises NoLevelFileError
        level_file_relpath = []
        if self.level_file_directory:
            level_file_relpath.append(self.level_file_directory)
        level_file_relpath.append(self.level_file)
        level_file_path = self.search.isfile(os.path.join(*level_file_relpath))
        if not level_file_path:
            raise NoLevelFileError(
                f"No level file in {os.path.join(self.directory, *level_file_relpath)}"
            )
        self.level_file_path = level_file_path


class SegmentLevel(Context):

    ACCOUNT_REQUIRED = True

    level_name = "segment"
    level_file = "segment.json"

    def setup(self):
        segment = self.search.basename()
        if self.search.isfile(EnvironmentLevel.level_file, up=1):
            self.cwd = self.search.parent(up=1)
            environment = self.search.basename(up=1)
        else:
            environment = segment
            segment = "default"
            self.cwd = self.search.parent(up=2)
        self.props["Environment"] = environment
        self.props["Segment"] = segment


class EnvironmentLevel(Context):

    ACCOUNT_REQUIRED = True

    level_name = "environment"
    level_file = "environment.json"

    def setup(self):
        self.props["Environment"] = self.search.basename()
        self.cwd = self.search.parent(up=2)


class RootLevel(Context):
    level_name = "root"
    level_file = "root.json"
    level_file_directory = None


class TenantLevel(Context):
    level_name = "tenant"
    level_file = "tenant.json"
    level_file_directory = None


class AccountLevel(Context):

    TENANT_REQUIRED = True

    level_name = "account"
    level_file = "account.json"
    level_file_directory = "config"

    def setup(self):
        self.props.update(File(self.level_file_path).load())


class ProductLevel(Context):

    ACCOUNT_REQUIRED = True

    level_name = "product"
    level_file = "product.json"
    level_file_directory = "config"

    def setup(self):
        self._product_dir = self.directory


class IntegratorLevel(Context):

    name = "integrator"
    filename = "integrator.json"

    def setup(self):
        self.props["Integrator"] = self.search.basename()



import click
from hamlet.backend.run import lambda_func as run_lambda_func_backend
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options


@click.command(
    "lambda",
    short_help="Run an AWS Lambda Function",
    context_settings=dict(max_content_width=240),
    deprecated=True,
)
@click.option(
    "-f",
    "--function-id",
    help="id of the function in the lambda deployment to run",
    required=True,
)
@click.option(
    "-u",
    "--deployment-unit",
    help="lambda deployment unit you want to execute",
    required=True,
)
@click.option(
    "-i", "--input-payload", help="json based payload you want to run the lambda with"
)
@click.option(
    "-l",
    "--include-log-tail",
    help="include the last 4kb of the execution log",
    is_flag=True,
)
@exceptions.backend_handler()
@pass_options
def lambda_func(options, **kwargs):
    """
    Run an AWS Lambda Function
    """

    args = {**options.opts, **kwargs}

    run_lambda_func_backend.run(**args, engine=options.engine, _is_cli=True)



import random
import string
from hamlet.backend.generate import utils
from . import template as account_cmdb_template


def generate_account_seed():
    seed = list(string.ascii_lowercase + string.digits)
    random.shuffle(seed)
    return "".join(seed[:10])


def run(**kwargs):
    utils.cookiecutter(account_cmdb_template, **kwargs)



# Here we controll modules loading order
# We can do it in command.__init__ but I guess this is a little bit more explicit
import hamlet.command  # noqa
import hamlet.command.autocomplete  # noqa
import hamlet.command.cmdb  # noqa
import hamlet.command.component  # noqa
import hamlet.command.deploy  # noqa
import hamlet.command.engine  # noqa
import hamlet.command.entrance  # noqa
import hamlet.command.generate  # noqa
import hamlet.command.layer  # noqa
import hamlet.command.manage  # noqa
import hamlet.command.reference  # noqa
import hamlet.command.release  # noqa
import hamlet.command.run  # noqa
import hamlet.command.schema  # noqa
import hamlet.command.setup  # noqa
import hamlet.command.task  # noqa
import hamlet.command.test  # noqa
import hamlet.command.visual  # noqa



import collections
import json
import os

from unittest import mock
from click.testing import CliRunner
from hamlet.command.deploy.test import test_deployments
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)

ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["-m,--deployment-mode"] = "DeploymentMode1"
ALL_VALID_OPTIONS["-l,--deployment-group"] = "DeploymentGroup1"
ALL_VALID_OPTIONS["-u,--deployment-unit"] = "DeploymentUnit1"
ALL_VALID_OPTIONS["-o,--output-dir"] = "output_dir"
ALL_VALID_OPTIONS["-p,--pytest-args"] = "PyTestArg1"
ALL_VALID_OPTIONS["-s,--silent"] = False


def template_backend_run_mock(data):
    def run(
        output_filename="unitlist-managementcontract.json",
        output_dir=None,
        *args,
        **kwargs
    ):
        os.makedirs(output_dir, exist_ok=True)
        unitlist_filename = os.path.join(output_dir, output_filename)
        with open(unitlist_filename, "wt+") as f:
            json.dump(data, f)

    return run


def mock_backend(unitlist=None):
    def decorator(func):
        @mock.patch("hamlet.command.deploy.test.test_run_backend")
        @mock.patch("hamlet.command.deploy.test.test_generate_backend")
        @mock.patch("hamlet.command.deploy.test.create_template_backend")
        @mock.patch("hamlet.command.deploy.test.create_deployment")
        @mock.patch("hamlet.backend.query.template")
        def wrapper(
            blueprint_mock,
            create_deployment_backend,
            create_template_backend,
            *args,
            **kwargs
        ):
            blueprint_mock.run.side_effect = template_backend_run_mock(unitlist)

            return func(blueprint_mock, create_template_backend, *args, **kwargs)

        return wrapper

    return decorator


unit_list = {
    "Stages": [
        {
            "Id": "StageId1",
            "Steps": [
                {
                    "Id": "StepId1",
                    "Parameters": {
                        "DeploymentUnit": "DeploymentUnit1",
                        "DeploymentGroup": "DeploymentGroup1",
                        "DeploymentProvider": "aws",
                        "Operations": ["Operation11"],
                        "CurrentState": "deployed",
                    },
                },
                {
                    "Id": "StepId2",
                    "Parameters": {
                        "DeploymentUnit": "DeploymentUnit2",
                        "DeploymentGroup": "DeploymentGroup2",
                        "DeploymentProvider": "aws",
                        "Operations": ["Operation21"],
                        "CurrentState": "deployed",
                    },
                },
            ],
        }
    ]
}


@mock_backend(unit_list)
def test_input_valid(
    blueprint_mock,
    create_template_backend,
    test_generate_backend,
    test_run_backend,
):

    run_options_test(
        CliRunner(), test_deployments, ALL_VALID_OPTIONS, blueprint_mock.run
    )


@mock_backend(unit_list)
def test_input_validation(
    blueprint_mock,
    create_template_backend,
    test_generate_backend,
    test_run_backend,
):

    runner = CliRunner()
    run_validatable_option_test(
        runner,
        test_deployments,
        create_template_backend.run,
        {
            "-m": "DeploymentMode1",
            "-l": "DeploymentGroup1",
            "-u": "DeploymentUnit1",
        },
        [],
    )



import os
import shutil
import json

from .exceptions import EngineStoreMissingEngineException

from .loaders.base import InstalledEngineLoader
from .loaders.shim import ShimEngineLoader, BundledShimEngineLoader
from .loaders.unicycle import UnicycleEngineLoader
from .loaders.tram import LatestTramEngineLoader, TramEngineLoader
from .loaders.train import LatestTrainEngineLoader, TrainEngineLoader
from .loaders.user import UserDefinedEngineLoader


class EngineStore:
    """
    The EngineStore manages the collection of engines that can be used for hamlet
    Loaders are assigned to the engine store which provide engines to the store
    It manages the location of the engine installations and maintains a persistant state
    that tracks global properties that effect all engines
    """

    def __init__(self, store_dir, config_search_paths):
        self._engines = []
        self.store_dir = store_dir
        self._store_state_file = os.path.join(self.store_dir, "store_state.json")
        self.engine_dir = os.path.join(self.store_dir, "engines")
        self.config_search_paths = config_search_paths

        self.store_state = {}
        self.load_store_state()

        self._default_engine = self.store_state.get(
            "default_engine", self.store_state.get("global_engine", None)
        )

        self._engine_locations = {
            "installed": {
                "loaded": False,
                "description": "Engines installed in the local engine store",
                "loaders": [
                    InstalledEngineLoader(engine_dir=self.engine_dir),
                ],
            },
            "local": {
                "loaded": False,
                "description": "Local user defined engines that are available",
                "loaders": [
                    ShimEngineLoader(),
                    BundledShimEngineLoader(),
                    UserDefinedEngineLoader(
                        config_search_paths=self.config_search_paths
                    ),
                ],
            },
            "remote": {
                "loaded": False,
                "description": "Engines available from remote locations",
                "loaders": [
                    UnicycleEngineLoader(),
                    LatestTramEngineLoader(),
                    LatestTrainEngineLoader(),
                    TrainEngineLoader(),
                ],
            },
            "hidden": {
                "loaded": False,
                "description": "Engines which are used internally or for release management",
                "loaders": [
                    TramEngineLoader(),
                ],
            },
        }

    @property
    def engine_locations(self):
        return self._engine_locations

    @property
    def default_engine(self):
        """
        The global engine defines a special engine that uses other engines to provide parts
        the global provides a consistent location that will link to other engine installations
        """
        return self._default_engine

    @default_engine.setter
    def default_engine(self, name):
        """
        Setting the default engine loads the engine and calls the  set_default_engine
        hook on all available engines
        """
        if name is None:
            self._default_engine = None
            self.store_state = {"default_engine": None}

        else:

            self.load_engines("installed")
            engine = self.get_engine(name, ["installed"])

            self.load_engines(["local"])

            for local_engine in self.get_engines("local"):
                local_engine.set_default_engine(engine, self)

            self._default_engine = engine.name
            self.store_state = {"default_engine": engine.name}

        self.save_store_state()

    def get_engines(self, locations=None):
        """
        Get availale engines
        """
        if locations is None:
            locations = self.engine_locations.keys()

        return [engine for engine in self._engines if engine.location in locations]

    def get_engine(self, name, locations=None):
        """
        Find an engine
        """
        if locations is None:
            locations = self.engine_locations.keys()

        try:
            result = [
                engine
                for engine in self.get_engines(locations=locations)
                if engine.name == name
            ][0]
        except IndexError:
            raise EngineStoreMissingEngineException(
                (
                    f"[!] Could not find the requested engine\n"
                    f"[!] - name: {name}\n"
                    f"[!] - locations: {locations}\n"
                    "[!] Run hamlet engine list-engines to find the engine"
                )
            )
        return result

    def load_engines(self, locations=None, refresh=False):
        if locations is None:
            locations = self.engine_locations.keys()

        if refresh:
            self._engines = []

        for location in locations:

            try:
                engine_location = self.engine_locations[location]
            except KeyError:
                continue

            if not engine_location["loaded"] or refresh:
                for loader in engine_location["loaders"]:
                    for engine in loader.load():
                        engine.engine_dir = self.engine_dir
                        engine.location = location
                        self._engines.append(engine)
                engine_location["loaded"] = True
        self.save_store_state()

    def load_store_state(self):
        """
        Load the persisted enginestore state
        """
        if os.path.isfile(self._store_state_file):
            with open(self._store_state_file, "r") as file:
                self.store_state = json.load(file)

    def save_store_state(self):
        """
        Save the persisted enginestore state
        """
        if self.store_state:
            if not os.path.isdir(self.store_dir):
                os.makedirs(self.store_dir)

            with open(self._store_state_file, "w") as file:
                json.dump(self.store_state, file)

    def clean_engines(self):
        if os.path.isdir(os.path.join(self.engine_dir)):
            shutil.rmtree(self.engine_dir)

    def clean_engine(self, name):
        if os.path.isdir(os.path.join(self.engine_dir, name)):
            shutil.rmtree(os.path.join(self.engine_dir, name))



import re
from hamlet.loggers import logging


logger = logging.getLogger("UTILS")


def deep_dict_update(a, b):
    for b_key, b_value in b.items():
        a_value = a.get(b_key)
        if isinstance(a_value, dict) and isinstance(b_value, dict):
            deep_dict_update(a_value, b_value)
        else:
            a[b_key] = b_value


# -- semver handling --
# Comparisons/naming roughly aligned to https://github.com/npm/node-semver
# in case we want to replace these routines with calls to this package via
# docker
def semver_valid(version):
    pattern = (
        r"^v?(0|[1-9][0-9]*)\.(0|[1-9][0-9]*)\.(0|[1-9][0-9]*)(\-([^+]+))?(\+(.*))?$"
    )
    match = re.search(pattern, version)
    if not match:
        raise ValueError(f'Invalid semantic version format "{version}"')
    major = match.group(1)
    minor = match.group(2)
    patch = match.group(3)
    prere = match.group(5)
    build = match.group(7)
    return (
        major,
        minor,
        patch,
        prere,
        build,
    )


# Strip any leading "v" (note we handle leading = in semver_satisfies)
# Convert any range indicators ("x" or "X") to 0
# * not supported as substitute for x
def semver_clean(version):
    # Handle the full format
    pattern = r"^v?(0|[1-9][0-9]*|x|X)\.(0|[1-9][0-9]*|x|X)\.(0|[1-9][0-9]*|x|X)(\-([^+]+))?(\+(.*))?$"
    match = re.search(pattern, version)
    if match:
        major = match.group(1).lower().replace("x", "0")
        minor = match.group(2).lower().replace("x", "0")
        patch = match.group(3).lower().replace("x", "0")
        prere = match.group(5)
        build = match.group(7)
        version = f"{major}.{minor}.{patch}"
        version = f"{version}-{prere}" if prere else version
        version = f"{version}+{build}" if build else version
        return version
    # Handle major.minor
    pattern = r"^v?(0|[1-9][0-9]*|x|X)\.(0|[1-9][0-9]*|x|X)$"
    match = re.search(pattern, version)
    if match:
        major = match.group(1).lower().replace("x", "0")
        minor = match.group(2).lower().replace("x", "0")
        return f"{major}.{minor}.0"
    # Handle major
    pattern = r"^v?(0|[1-9][0-9]*|x|X)$"
    match = re.search(pattern, version)
    if match:
        major = match.group(1).lower().replace("x", "0")
        return f"{major}.0.0"
    raise ValueError(f'Invalid semantic version format "{version}"')


def semver_compare(v1, v2):

    v1 = semver_clean(v1)
    v2 = semver_clean(v2)

    v1_components = semver_valid(v1)
    v2_components = semver_valid(v2)

    # MAJOR, MINOR and PATCH should compare numericaly
    for i in range(3):
        if v1_components[i] > v2_components[i]:
            return 1
        elif v1_components[i] < v2_components[i]:
            return -1

    # PREREL, BUILD should compare with the ASCII order.
    for i in range(3, 5):
        if not v1_components[i] and not v2_components[i]:
            continue
        if not v1_components[i] and v2_components[i]:
            return -1
        if v1_components[i] and not v2_components[i]:
            return 1
        if v1_components[i] < v2_components[i]:
            return -1
        else:
            return 1
    return 0


# a comparator set is a list of comparators, true if all comparators are true
# a comparator is an operator and a version
def semver_satisfies(version, comparators_set):
    for comparator in comparators_set:
        logger.debug('Checking comparator "%s" ...', comparator)
        match = re.match(r"^(<=|>=|=|<|>)(.+)$", comparator)
        if not match:
            raise ValueError(f"Unknown comparator {comparator}")
        operator = match.group(1)
        comparator_version = semver_clean(match.group(2))
        comparator_result = semver_compare(version, comparator_version)
        if operator == "<" and comparator_result < 0:
            continue
        elif operator == ">" and comparator_result > 0:
            continue
        elif operator == "=" and comparator_result == 0:
            continue
        elif operator == "<=" and comparator_result <= 0:
            continue
        elif operator == ">=" and comparator_result >= 0:
            continue
        else:
            return False
    return True


def semver_upgrade_list(upgrade_list, maximum_version):
    # assuming the list is ordered
    for i in range(len(upgrade_list)):
        if semver_compare(maximum_version, upgrade_list[i]) < 0:
            return upgrade_list[:i]
    return upgrade_list



# NOTE: this file must remain valid python file in order to perform tests on it


def checkov_test(filename, framework, skip_checks=None):
    import json
    import subprocess
    import shutil

    if shutil.which("checkov") is None:
        raise Exception(
            "checkov command not found install from https://www.checkov.io/"
        )

    cmd_args = [
        "checkov",
        "--output",
        "json",
        "--framework",
        framework,
        "--quiet",
        "-f",
        filename,
    ]

    if skip_checks is not None:
        cmd_args.append("--skip-check")
        cmd_args.append(",".join(skip_checks))

    result = subprocess.run(
        cmd_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf8"
    )

    if result.returncode == 1:
        result = json.loads(result.stdout)
        if result["summary"]["failed"] >= 0:
            failure_details = []
            for detail in result["results"]["failed_checks"]:
                failure_details.append(
                    {
                        k: v
                        for k, v in detail.items()
                        if k
                        in [
                            "check_name",
                            "check_id",
                            "check_class",
                            "resource",
                            "guideline",
                            "check_result",
                        ]
                    }
                )
            raise AssertionError(json.dumps(failure_details, indent=4))

    if result.returncode >= 1:
        raise Exception(result.stderr)



import click
import os
import subprocess
from shutil import copyfile

from hamlet.command import root as cli


def regenerate_autocompletion_files(shell_type: str, autocomplete_path: str) -> None:

    subprocess.check_call(
        f"_HAMLET_COMPLETE={shell_type}_source hamlet > {autocomplete_path}",
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )
    print(f"Regenerated {shell_type} completion file: {autocomplete_path}")


def autocomplete_fish(user_home_path: str, autocomplete_path: str) -> None:
    destination_path = os.path.join(
        user_home_path, ".config/fish/completions/hamlet.fish"
    )

    if os.path.isfile(destination_path):
        print("Autocompletion was previously set up. Skipping.")

    else:
        copyfile(
            autocomplete_path,
            destination_path,
        )


def autocomplete_bash_or_zsh(
    user_home_path: str, autocomplete_path: str, shell_type: str = "bash"
) -> None:

    autocomplete_path = autocomplete_path.replace(user_home_path, "~")

    autocompletion_string = f"# HAMLET-AUTOCOMPLETE\n. {autocomplete_path}"
    run_command_path = os.path.join(user_home_path, f".{shell_type}rc")

    with open(run_command_path, "r") as OH:
        if autocompletion_string in OH.readlines():
            print("Autocompletion was previously set up. Skipping.")

        else:
            with open(run_command_path, "a+") as OH:
                OH.write(f"\n{autocompletion_string}")


@cli.command()
@click.argument(
    "shell-type",
    type=click.Choice(["bash", "zsh", "fish"], case_sensitive=False),
    default="bash",
)
@click.option(
    "--regenerate",
    help="regenerates the autocompletion file",
    type=click.BOOL,
    default=False,
    is_flag=True,
)
def autocomplete(shell_type: str, regenerate: bool) -> None:
    """
    Enable autocomplete for the cli
    """
    user_home_path = os.path.expanduser("~")
    autocomplete_path = os.path.join(
        __path__[0], "scripts", f".hamlet-complete.{shell_type}"
    )

    if regenerate:
        regenerate_autocompletion_files(shell_type, autocomplete_path)

    if shell_type in ("bash", "zsh"):
        autocomplete_bash_or_zsh(user_home_path, autocomplete_path, shell_type)
    elif "fish" == shell_type:
        autocomplete_fish(user_home_path, autocomplete_path)

    click.secho(
        f"Autocomplete setup for {shell_type} open a new shell to complete the process",
        color="green",
    )



import click
from hamlet.backend.manage import crypto as manage_crypto_backend
from hamlet.command.common.config import pass_options
from hamlet.command.common import exceptions


@click.command(
    "crypto",
    short_help="Manage cryptographic operations using KMS",
    context_settings=dict(max_content_width=240),
    deprecated=True,
)
@click.option("-a", "--alias", help="alias for the master key to be used")
@click.option(
    "-b",
    "--base64-decode",
    is_flag=True,
    help="force base64 decode of the input before processing",
)
@click.option("-d", "--decrypt", is_flag=True, help="decrypt operation")
@click.option("-e", "--encrypt", is_flag=True, help="encrypt operation")
@click.option(
    "-f",
    "--crypto-file",
    help="file which contains the plaintext or ciphertext to be processed",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
)
@click.option("-k", "--key-id", help="master key to be used")
@click.option(
    "-n",
    "--no-alteration",
    help="no alteration to CRYPTO_TEXT (pass through as is)",
    is_flag=True,
)
@click.option(
    "-p", "--json-path", help="path to the attribute within CRYPTO_FILE to be processed"
)
@click.option("-q", "--quiet", is_flag=True, help="don't display result(quiet)")
@click.option("-t", "--crypto-text", help="plaintext or ciphertext to be processed")
@click.option(
    "-u",
    "--update",
    is_flag=True,
    help="update the attribute at JSON_PATH (if provided), or replace CRYPTO_FILE with operation result",
)
@click.option(
    "-v", "--visible", help="result is base64 decoded (visible)", is_flag=True
)
@exceptions.backend_handler()
@pass_options
def crypto(options, **kwargs):
    """
    Manage cryptographic operations using KMS

    \b
    NOTES:
    1. If a file is required but not provided, the default filename
         will be expected in the equivalent directory of the infrastructure tree
    2. If JSON_PATH is provided,
       - a CRYPTO_FILE is required
       - the targetted file must be JSON format
       - encrypt requires CRYPTO_TEXT to be provided, or for the attribute to
         to present
       - attribute is updated with the operation result if update flag is set
    3. If JSON_PATH is NOT provided,
       - one of CRYPTO_FILE or CRYPTO_TEXT must be provided
       - CRYPTO_TEXT takes precedence over CRYPTO_FILE
    4. If a file at CRYPTO_FILE can't be located based on current directory, it will be
       treated as a relative directory using the default filename
    5. Don't include "alias/" in any provided alias
    6. If encrypting, the key is located as follows,
       - use KEYID if provided
       - use ALIAS if provided
       - if in segment directory, use segment keyid if available
       - if in product directory, use product keyid if available
       - if in account directory, use account keyid if available
       - otherwise error
    7. The result is sent to stdout and is base64 encoded unless the
       visibility flag is set
    8. Decrypted files will have a ".decrypted" extension added so they can be ignored by git
    """

    args = {**options.opts, **kwargs}

    manage_crypto_backend.run(**args, engine=options.engine, _is_cli=True)



import boto3
import json
import jmespath


def run(
    SecretArn,
    JSONKeyPath,
    Region=None,
    AWSAccessKeyId=None,
    AWSSecretAccessKey=None,
    AWSSessionToken=None,
    env={},
):
    """
    Retrieve the value of secrets manager secret
    """

    session = boto3.Session(
        aws_access_key_id=AWSAccessKeyId,
        aws_secret_access_key=AWSSecretAccessKey,
        aws_session_token=AWSSessionToken,
        region_name=Region,
    )

    secretsmanager = session.client("secretsmanager")
    secret_response = secretsmanager.get_secret_value(SecretId=SecretArn)

    try:
        secret_value = json.loads(secret_response["SecretString"])
        secret_value = jmespath.search(JSONKeyPath, secret_value)

    except json.JSONDecodeError:
        secret_value = secret_response["SecretString"]

    return {"Properties": {"secret_value": secret_value}}



import os
from jinja2 import Environment, FileSystemLoader

TEMPLATES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")


loader = Environment(
    loader=FileSystemLoader(TEMPLATES_DIR),
    trim_blocks=True,
    lstrip_blocks=True,
    autoescape=True,
)



import click
from hamlet.backend.run import expo_app_publish as run_expo_app_publish_backend
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options


@click.command(
    "expo-app-publish",
    short_help="Run a task based build of an Expo app binary",
    context_settings=dict(max_content_width=240),
)
@click.option(
    "-l",
    "--deployment-group",
    help="mobile app deployment group",
    default="application",
    envvar="DEPLOYMENT_GROUP",
)
@click.option(
    "-u",
    "--deployment-unit",
    required=True,
    help="mobile app deployment unit",
    envvar="DEPLOYMENT_UNIT",
)
@click.option(
    "-n",
    "--node-package-manager",
    help="The node package manager to use for builds",
    type=click.Choice(["auto", "yarn", "npm"]),
    envvar="NODE_PACKAGE_MANAGER",
)
@click.option(
    "-o",
    "--binary-output-dir",
    type=click.Path(dir_okay=True, file_okay=False, writable=True, resolve_path=True),
    help="The directory to save generated binaries to",
)
@exceptions.backend_handler()
@pass_options
def expo_app_publish(options, **kwargs):
    """
    Run a task based build of an Expo app binary
    """

    args = {**options.opts, **kwargs}
    run_expo_app_publish_backend.run(
        **args,
        build_logs=(options.opts.get("log_level", "info") in ["debug", "trace"]),
        engine=options.engine,
        _is_cli=True
    )



from hamlet.backend.common.exceptions import BackendException

import docker
import docker.errors


def run(Image, env={}):
    """
    pull a docker image to the local docker host
    """

    client = docker.from_env()

    try:
        client.ping()
    except docker.errors.APIError as e:
        raise BackendException(str(e))

    try:
        client.images.pull(repository=Image, all_tags=False)
    except docker.errors.NotFound as e:
        raise BackendException(str(e))

    return {"Properties": {}}



import click

from hamlet.command.common import decorators

try:
    from hamlet.__version__ import version
except ImportError:
    version = "unknown"


@click.group("root", context_settings=dict(max_content_width=240))
@click.version_option(version)
@decorators.common_cli_config_options
@decorators.common_generation_options
@decorators.common_logging_options
@decorators.common_engine_options
@decorators.common_district_options
@click.pass_context
def root(ctx, opts):
    """
    hamlet
    """
    pass



import shutil
import subprocess


class EngineCodeSourceBuildData:
    """
    Provides utility functions that generate version details about
    an engine source as part of its build process
    """

    def _git_path(self):
        return shutil.which("git")

    def __init__(self, path):
        self.path = path

    def _is_git_repo(self):
        try:
            subprocess.check_output(
                [
                    self._git_path(),
                    "rev-parse",
                    "--is-inside-work-tree",
                ],
                cwd=self.path,
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError:
            return False

        return True

    def _git_current_revision(self):
        return subprocess.check_output(
            [
                self._git_path(),
                "rev-parse",
                "HEAD",
            ],
            cwd=self.path,
            text=True,
        ).strip()

    def _git_current_tag(self):
        return subprocess.check_output(
            [
                self._git_path(),
                "tag",
                "--points-at",
                "HEAD",
            ],
            cwd=self.path,
            text=True,
        ).strip()

    def _git_current_origin(self):
        return subprocess.check_output(
            [
                self._git_path(),
                "remote",
                "get-url",
                "origin",
            ],
            cwd=self.path,
            text=True,
        ).strip()

    @property
    def details(self):

        details = {}

        if self._is_git_repo():
            details["code_source"] = {
                "provider": "git",
                "revision": self._git_current_revision(),
                "tag": self._git_current_tag(),
                "origin": self._git_current_origin(),
            }

        return details



import json
import boto3

from botocore.exceptions import WaiterError
from hamlet.backend.common.utilities import to_bool
from hamlet.backend.common.exceptions import BackendException


def run(
    ClusterArn=None,
    TaskFamily=None,
    ContainerName=None,
    CapacityProvider=None,
    SubnetIds=None,
    SecurityGroupIds=None,
    PublicIP=None,
    CommandOverride=None,
    EnvironmentOverrides=None,
    Region=None,
    AWSAccessKeyId=None,
    AWSSecretAccessKey=None,
    AWSSessionToken=None,
    env={},
):
    """
    Start a new task using the provided TaskFamily
    with any overrides applied to a given container
    """

    session = boto3.Session(
        aws_access_key_id=AWSAccessKeyId,
        aws_secret_access_key=AWSSecretAccessKey,
        aws_session_token=AWSSessionToken,
        region_name=Region,
    )
    ecs = session.client("ecs")

    run_task_args = {
        "cluster": ClusterArn,
        "taskDefinition": TaskFamily,
        "capacityProviderStrategy": [
            {
                "capacityProvider": CapacityProvider,
            }
        ],
        "count": 1,
        "enableECSManagedTags": True,
        "propagateTags": "TASK_DEFINITION",
    }

    if SubnetIds and SecurityGroupIds:
        run_task_args["networkConfiguration"] = {
            "awsvpcConfiguration": {
                "subnets": SubnetIds.split(","),
                "securityGroups": SecurityGroupIds.split(","),
                "assignPublicIp": "ENABLED" if to_bool(PublicIP, False) else "DISABLED",
            }
        }

    if CommandOverride or EnvironmentOverrides:
        container_override = {"name": ContainerName}

        if CommandOverride:

            try:
                command = json.loads(CommandOverride)
            except json.decoder.JSONDecodeError:
                command = [CommandOverride]

            if isinstance(command, str):
                command = [command.split(" ")]

            container_override["command"] = command

        if EnvironmentOverrides:
            container_override["environment"] = [
                {"name": key, "value": value}
                for key, value in json.loads(EnvironmentOverrides).items()
            ]

        run_task_args["overrides"] = {"containerOverrides": [container_override]}

    task_response = ecs.run_task(**run_task_args)

    task_args = {
        "cluster": ClusterArn,
        "tasks": [task["taskArn"] for task in task_response["tasks"]],
    }

    try:
        running_waiter = ecs.get_waiter("tasks_running")
        running_waiter.wait(**task_args)

    except WaiterError:
        if True in [
            True
            for task in ecs.describe_tasks(**task_args)["tasks"]
            if task["lastStatus"] == "STOPPED"
        ]:
            pass

    while True:
        try:
            stopped_waiter = ecs.get_waiter("tasks_stopped")
            stopped_waiter.wait(**task_args)
        except WaiterError:
            continue

        break

    task_describe = ecs.describe_tasks(**task_args)
    run_status = {"tasks": []}

    if task_describe.get("failures", None):
        raise BackendException(task_describe["failures"])

    task_keys = [
        "stopCode",
        "stoppedReason",
        "taskArn",
        "lastStatus",
        "capacityProviderName",
    ]
    container_keys = [
        "name",
        "image",
        "runtimeId",
        "lastStatus",
        "exitCode",
        "reason",
    ]

    for task in task_describe["tasks"]:
        task_status = {k: v for k, v in task.items() if k in task_keys}

        task_status["startedAt"] = task["startedAt"].isoformat()
        task_status["stoppedAt"] = task["stoppedAt"].isoformat()

        task_status["containers"] = []
        for container in task["containers"]:
            task_status["containers"].append(
                {k: v for k, v in container.items() if k in container_keys}
            )
            if container["name"] == ContainerName and container["exitCode"] != 0:
                raise BackendException(
                    (
                        "Primary container did not run successfully "
                        f"- exitCode: {container['exitCode']}"
                    )
                )

        run_status["tasks"].append(task_status)

    return {
        "Properties": {
            "task_arns": json.dumps(
                [task["taskArn"] for task in task_response["tasks"]]
            ),
            "run_status": json.dumps(run_status),
        }
    }



import json
import click


def run(Value, Format, OutputStream, env={}):
    """
    echo a given value to an output stream
    """

    err = OutputStream == "stderr"
    if Format == "json":
        Value = json.dumps(json.loads(Value), indent=2)

    click.echo(Value, err=err)
    return {"Properties": {"result": Value}}



import os
import json
import tempfile

import jmespath

from jmespath.exceptions import JMESPathError
from hamlet.backend.create import template
from hamlet.backend.common import exceptions


def run(
    deployment_mode=None,
    generation_entrance=None,
    generation_entrance_parameter=None,
    generation_input_source=None,
    generation_provider=None,
    generation_framework=None,
    output_filename=None,
    query_text=None,
    query_params=None,
    sub_query_text=None,
    log_level=None,
    root_dir=None,
    district_type=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    engine=None,
    **kwargs,
):
    query = Query(
        deployment_mode=deployment_mode,
        generation_entrance=generation_entrance,
        generation_entrance_parameter=generation_entrance_parameter,
        generation_input_source=generation_input_source,
        generation_provider=generation_provider,
        generation_framework=generation_framework,
        output_filename=output_filename,
        log_level=log_level,
        root_dir=root_dir,
        district_type=district_type,
        tenant=tenant,
        account=account,
        product=product,
        environment=environment,
        segment=segment,
        engine=engine,
    )
    if query_text is None:
        result = query.blueprint_data
    else:
        result = query.query(query_text, query_params or {})
    if sub_query_text is not None:
        result = query.perform_query(sub_query_text, result)
    return result


class Query:
    def __init__(
        self,
        deployment_mode,
        generation_entrance=None,
        generation_entrance_parameter=None,
        generation_input_source=None,
        generation_provider=None,
        generation_framework=None,
        output_filename=None,
        log_level=None,
        root_dir=None,
        district_type=None,
        tenant=None,
        account=None,
        product=None,
        environment=None,
        segment=None,
        engine=None,
    ):
        with tempfile.TemporaryDirectory() as query_dir:
            # mocked blueprint doesn't need the valid context
            output_filepath = os.path.join(query_dir, output_filename)
            template.run(
                output_dir=query_dir,
                deployment_mode=deployment_mode,
                entrance=generation_entrance,
                entrance_parameter=generation_entrance_parameter,
                generation_input_source=generation_input_source,
                generation_provider=generation_provider,
                generation_framework=generation_framework,
                log_level=log_level,
                root_dir=root_dir,
                district_type=district_type,
                tenant=tenant,
                account=account,
                product=product,
                environment=environment,
                segment=segment,
                engine=engine,
            )

            try:
                with open(output_filepath, "rt") as f:
                    self.blueprint_data = json.load(f)
            except FileNotFoundError:
                raise exceptions.BackendException(
                    f"Query using entrance {generation_entrance} didn't return any results"
                )

    def query(self, query, params=None, require=None):
        require = require or []
        params = params or {}
        if require:
            for key in require:
                try:
                    params[key]
                except KeyError as e:
                    raise exceptions.BackendException(
                        f'Missing required query param:"{key}".'
                    ) from e
        if params:
            # jsonify every param
            for key, value in params.items():
                jsonified_value = json.dumps(value)
                # need to replace double quotes with single quotes because jq is a really weird
                # and works correctly only with single quotes
                if jsonified_value.startswith('"') and jsonified_value.endswith('"'):
                    jsonified_value = f"'{jsonified_value[1:-1]}'"
                params[key] = jsonified_value
            query = query.format(**params)

        return self.perform_query(query, self.blueprint_data)

    def perform_query(self, query, data):
        if not query:
            raise exceptions.BackendException("Query can not be empty")
        try:
            return jmespath.search(query, data)
        except JMESPathError as e:
            raise exceptions.BackendException(f"JMESPath query error: {str(e)}") from e



import boto3
from base64 import b64decode


def run(
    Ciphertext,
    EncryptionScheme=None,
    Region=None,
    AWSAccessKeyId=None,
    AWSSecretAccessKey=None,
    AWSSessionToken=None,
    env={},
):
    """
    Given a base64 encoded aws kms ciphertext blob decrypt it using
    the provided credentials and return the plaintext result as a string
    """

    if EncryptionScheme:
        Ciphertext = Ciphertext.replace(EncryptionScheme, "", 1)
    Ciphertext = b64decode(Ciphertext)

    session = boto3.Session(
        aws_access_key_id=AWSAccessKeyId,
        aws_secret_access_key=AWSSecretAccessKey,
        aws_session_token=AWSSessionToken,
        region_name=Region,
    )

    kms = session.client("kms")
    result = kms.decrypt(
        CiphertextBlob=Ciphertext,
    )

    return {"Properties": {"result": result["Plaintext"].decode("utf-8")}}



import boto3
import os


def run(
    BucketName,
    Prefix,
    LocalPath,
    Region=None,
    AWSAccessKeyId=None,
    AWSSecretAccessKey=None,
    AWSSessionToken=None,
    env={},
):
    """
    Download the objects matching a prefix to a local directory from an S3 Bucket
    """

    session = boto3.Session(
        aws_access_key_id=AWSAccessKeyId,
        aws_secret_access_key=AWSSecretAccessKey,
        aws_session_token=AWSSessionToken,
        region_name=Region,
    )

    s3 = session.resource("s3")
    bucket = s3.Bucket(BucketName)
    object_count = 0

    for object in bucket.objects.filter(Prefix=Prefix):
        object_count += 1
        local_path = os.path.join(
            os.path.abspath(LocalPath),
            object.key[len(os.path.commonpath([Prefix, object.key])) :],
        )

        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        s3.Object(object.bucket_name, object.key).download_file(local_path)

    return {
        "Properties": {
            "object_count": str(object_count),
            "s3_path": f"s3://{BucketName}/" + Prefix,
            "local_path": LocalPath,
        }
    }



import os
import json
import shutil
import subprocess

from .exceptions import BackendException

SHELL_SCRIPT_FILENAME = os.path.join(os.path.dirname(__file__), "shell_vars_to_json.sh")


def shell_vars_to_json(filename=None, vars_mapping=None):

    if shutil.which("bash") is None:
        raise BackendException("Could not find bash installation")

    args = [SHELL_SCRIPT_FILENAME, filename]
    for variable_name, variable_type in vars_mapping.items():
        args += [variable_name, variable_type]
    try:
        process = subprocess.Popen(
            [shutil.which("bash"), *args],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
        )
        process.wait()
        stderr = process.stderr.read()
        stdout = process.stdout.read()
        if stderr:
            raise RuntimeError(stderr)
        return json.loads(stdout)
    finally:
        try:
            process.kill()
        except ProcessLookupError:
            pass



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.run.sentry_release import sentry_release as run_sentry_release
from tests.unit.command.test_option_generation import run_options_test

ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["!-u,--deployment-unit"] = "deployment_unit"
ALL_VALID_OPTIONS["!-m,--sentry-source-map-s3-url"] = "url"
ALL_VALID_OPTIONS["!-r,--sentry-release-name"] = "release_name"
ALL_VALID_OPTIONS["-g,--deployment-group"] = "deployment_group"
ALL_VALID_OPTIONS["-p,--sentry-url-prefix"] = "url_prefix"
ALL_VALID_OPTIONS["-a,--app-type"] = "react-native"
ALL_VALID_OPTIONS["-d,--dist"] = "distribution"


@mock.patch("hamlet.command.run.sentry_release.run_sentry_release_backend")
def test_input_valid(run_sentry_release_backend):
    run_options_test(
        CliRunner(),
        run_sentry_release,
        ALL_VALID_OPTIONS,
        run_sentry_release_backend.run,
    )



from hamlet.backend.common.exceptions import BackendException
import boto3


def run(
    FunctionArn,
    Payload,
    Region=None,
    AWSAccessKeyId=None,
    AWSSecretAccessKey=None,
    AWSSessionToken=None,
    env={},
):
    """
    Invoke a lambda function synchronously and wait for the result
    Fail on lambda error and return logs and result
    """

    session = boto3.Session(
        aws_access_key_id=AWSAccessKeyId,
        aws_secret_access_key=AWSSecretAccessKey,
        aws_session_token=AWSSessionToken,
        region_name=Region,
    )

    lambda_client = session.client("lambda")
    result = lambda_client.invoke(
        FunctionName=FunctionArn,
        InvocationType="RequestResponse",
        LogType="None",
        Payload=bytes(Payload, "utf-8"),
    )

    if "FunctionError" in result:
        raise BackendException(
            f"Lambda Function Failure - {result['FunctionError']} - {result['Payload'].read().decode('utf-8')}"
        )

    return {"Properties": {"payload": result["Payload"].read().decode("utf-8")}}



import os

from hamlet.backend import query as query_backend

GET_PROPERTIES_ENVIRONMENT_QUERY = "Environment"


def get_automation_properties(engine=None, **kwargs):

    query_args = {
        **kwargs,
        "generation_entrance": "releaseinfo",
        "output_filename": "releaseinfo-config.json",
        "output_format": "default",
    }
    automation_properties = query_backend.run(
        **query_args,
        cwd=os.getcwd(),
        query_text=GET_PROPERTIES_ENVIRONMENT_QUERY,
        engine=engine
    )

    return automation_properties



import json
from unittest import mock
from click.testing import CliRunner

from hamlet.command.engine import (
    list_engines,
    describe_engine,
    clean_engines,
    install_engine,
    set_engine,
    env,
)


def test_list_engines(options):
    cli = CliRunner()
    result = cli.invoke(list_engines, ["--output-format", "json"], obj=options)
    assert result.exit_code == 0
    print(result.output)
    result = json.loads(result.output)
    assert len(result) == 1
    assert {
        "name": "Name[1]",
        "description": "Description[1]",
        "digest": "Digest[1]",
        "short_digest": "Digest[1]",
        "location": "local",
    } in result


@mock.patch("json.dumps", mock.MagicMock(return_value="{}"))
def test_describe_engine(options):
    cli = CliRunner()
    result = cli.invoke(describe_engine, obj=options)
    print(result.exc_info)
    assert result.exit_code == 0


def test_install_engine(options):
    cli = CliRunner()
    result = cli.invoke(install_engine, ["engine[1]"], obj=options)
    print(result.exception)
    assert result.exit_code == 0


def test_env(options):
    cli = CliRunner()
    result = cli.invoke(env, [], obj=options)
    print(result.exception)
    assert result.exit_code == 0
    print(result.output)
    assert result.output == "# run eval $(hamlet engine env) to set variables\n"


def test_set_engine(options):
    cli = CliRunner()
    result = cli.invoke(set_engine, ["Name[1]"], obj=options)
    print(result.exception)
    assert result.exit_code == 0


def test_clean_engines(options):
    cli = CliRunner()
    result = cli.invoke(clean_engines, [], obj=options)
    print(result.exception)
    assert result.exit_code == 0



from hamlet.command import root as cli
from hamlet.backend import setup as setup_backend
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options


@cli.command("setup", short_help="", context_settings=dict(max_content_width=240))
@exceptions.backend_handler()
@pass_options
def setup(options, **kwargs):
    """
    Loads the plugins defined as part of the current district
    """
    args = {**options.opts, **kwargs}
    setup_backend.run(**args, engine=options.engine, _is_cli=True)



from hamlet.backend.contract.tasks.exceptions import TaskConditionFailureException


def run(Condition, Test, Value, env={}):
    """
    Conditional testing service to handle skipping steps in a contract
    """
    result = False

    if Condition == "Equals":
        result = Test == Value
    elif Condition == "NotEquals":
        result = Test != Value
    elif Condition == "StartsWith":
        result = Value.startswith(Test)
    elif Condition == "EndsWith":
        result = Value.endswith(Test)
    elif Condition == "Contains":
        result = Value.contains(Test)

    if not result:
        raise TaskConditionFailureException("Condition not met")

    return {"Properties": {"result": result}}



import os
import json
import collections

from unittest import mock
from click.testing import CliRunner

from hamlet.command.reference import list_reference_types, list_references
from hamlet.command.common.config import Options
from tests.unit.command.test_option_generation import (
    run_options_test,
)


info_mock_output = {
    "ReferenceTypes": [
        {
            "Type": "ReferenceType[1]",
            "Description": "ReferenceDescription[1]",
            "Attributes": [{"Names": "Attribute[1]", "Types": ["string"]}],
            "PluralType": "ReferencePluralType[1]",
        },
        {
            "Type": "ReferenceType[2]",
            "Description": "ReferenceDescription[2]",
            "Attributes": [{"Names": "Attribute[2]", "Types": ["string"]}],
            "PluralType": "ReferencePluralType[2]",
        },
    ],
    "ReferenceData": [
        {
            "Type": "ReferenceType[1]",
            "Id": "ReferenceDataId[1]",
            "Properties": {"ReferenceProperty[1]": "ReferencePropertyValue[1]"},
        },
        {
            "Type": "ReferenceType[2]",
            "Id": "ReferenceDataId[2]",
            "Properties": {"ReferenceProperty[2]": "ReferencePropertyValue[2]"},
        },
    ],
}


def template_backend_run_mock(data):
    def run(output_filename="info.json", output_dir=None, *args, **kwargs):
        os.makedirs(output_dir, exist_ok=True)
        filename = os.path.join(output_dir, output_filename)
        with open(filename, "wt+") as f:
            json.dump(data, f)

    return run


def mock_backend(data=None):
    def decorator(func):
        @mock.patch("hamlet.backend.query.template")
        def wrapper(blueprint_mock, *args, **kwargs):
            blueprint_mock.run.side_effect = template_backend_run_mock(data)
            return func(blueprint_mock, *args, **kwargs)

        return wrapper

    return decorator


LIST_TYPES_VALID_OPTIONS = collections.OrderedDict()
LIST_TYPES_VALID_OPTIONS["-q,--query"] = "[]"
LIST_TYPES_VALID_OPTIONS["--output-format"] = "json"


@mock_backend(info_mock_output)
def test_list_reference_types_input_valid(blueprint_mock):
    run_options_test(
        CliRunner(), list_reference_types, LIST_TYPES_VALID_OPTIONS, blueprint_mock.run
    )


@mock_backend(info_mock_output)
def test_list_reference_types(blueprint_mock):
    obj = Options()

    cli = CliRunner()
    result = cli.invoke(list_reference_types, ["--output-format", "json"], obj=obj)
    print(result.exc_info)
    assert result.exit_code == 0
    result = json.loads(result.output)
    assert len(result) == 2
    assert {
        "Type": "ReferenceType[1]",
        "PluralType": "ReferencePluralType[1]",
        "Description": "ReferenceDescription[1]",
    } in result
    assert {
        "Type": "ReferenceType[2]",
        "PluralType": "ReferencePluralType[2]",
        "Description": "ReferenceDescription[2]",
    } in result


LIST_REFERENCES_VALID_OPTIONS = collections.OrderedDict()
LIST_REFERENCES_VALID_OPTIONS["-q,--query"] = "[]"
LIST_REFERENCES_VALID_OPTIONS["--output-format"] = "json"


@mock_backend(info_mock_output)
def test_list_references_input_valid(
    blueprint_mock,
):
    run_options_test(
        CliRunner(), list_references, LIST_REFERENCES_VALID_OPTIONS, blueprint_mock.run
    )


@mock_backend(info_mock_output)
def test_list_references(blueprint_mock):
    obj = Options()

    cli = CliRunner()
    result = cli.invoke(list_references, ["--output-format", "json"], obj=obj)
    print(result.exc_info)
    assert result.exit_code == 0
    result = json.loads(result.output)
    assert len(result) == 2
    assert {"Id": "ReferenceDataId[1]", "Type": "ReferenceType[1]"} in result
    assert {"Id": "ReferenceDataId[2]", "Type": "ReferenceType[2]"} in result



import collections
from unittest import mock
from click.testing import CliRunner

from hamlet.command.release import transfer_image as transfer_image

from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)


ALL_VALID_OPTIONS = collections.OrderedDict()

ALL_VALID_OPTIONS["!-u,--deployment-unit"] = "deployment_unit"
ALL_VALID_OPTIONS["!-r,--build-reference"] = "build_reference"
ALL_VALID_OPTIONS["--code-tag"] = "code_tag"
ALL_VALID_OPTIONS["-f,--image-format"] = "docker"
ALL_VALID_OPTIONS["-s,--registry-scope"] = "registry_scope"

ALL_VALID_OPTIONS["!--source-account"] = "source_account"
ALL_VALID_OPTIONS["!--source-environment"] = "source_environment"


@mock.patch(
    "hamlet.command.release.transfer_image_backend.TransferImageAutomationRunner"
)
def test_input_valid(transfer_image_backend):
    run_options_test(
        CliRunner(), transfer_image, ALL_VALID_OPTIONS, transfer_image_backend
    )


@mock.patch(
    "hamlet.command.release.transfer_image_backend.TransferImageAutomationRunner"
)
def test_input_validation(transfer_image_backend):
    runner = CliRunner()
    run_validatable_option_test(
        runner,
        transfer_image,
        transfer_image_backend,
        {
            "-u": "deployment_unit",
            "-r": "build_reference",
            "-f": "image_format",
            "--source-account": "source_account",
            "--source-environment": "source_environment",
        },
        [],
    )



from hamlet.loggers import logging
from hamlet.utils import DynamicOption

# TEST_OPTIONS = collections.OrderedDict()
# TEST_OPTIONS['!-i,--id,id'] = ('id')
# TEST_OPTIONS['--name,name'] = ('name', 'id')
# TEST_OPTIONS['--param,param'] = ('non-default', 'default')

logger = logging.getLogger("test-generate")


def __inputlines(*args):
    return "\n".join(list(str(a) for a in args))


def __split_key(key):
    key = key.strip("! ")
    parts = tuple(part.strip() for part in key.split(",") if part.strip())
    if len(parts) < 2:
        raise ValueError('Key should be in format:"--option-key, param_name"')
    return parts


def __get_key_part(key, index):
    parts = __split_key(key)
    return parts[max(len(parts) - 2, index)]


def __get_param_name_from_key(key):
    return __split_key(key)[-1]


def __is_required_key(key):
    return key.strip().startswith("!")


def __get_value(values):
    if isinstance(values, (list, tuple)):
        return values[0]
    else:
        return values


def __get_default(values):
    if isinstance(values, (list, tuple)):
        return values[1]
    raise ValueError("default value not set")


def __generate_prompt_options_testcase(template):
    for required_only in [True, False]:
        options = []
        options_str = []
        input = []
        expected = {}
        for keys, values in template.items():
            required = __is_required_key(keys)
            param = __get_param_name_from_key(keys)
            value = __get_value(values)
            if not required_only or required:
                if isinstance(value, bool):
                    input.append("y" if value else "n")
                else:
                    input.append(str(value))
            expected[param] = (
                value if not required_only or required else __get_default(values)
            )
        yield {
            "options": options,
            "options_str": options_str,
            "input": __inputlines(*input, "y"),
            "expected": expected,
        }


def __generate_options_testcase(template):
    for required_only in [True, False]:
        max_key_parts = 1
        for key in template:
            key_parts = __split_key(key)
            max_key_parts = max(max_key_parts, len(key_parts) - 1)
        for key_index in range(max_key_parts):
            options = []
            options_str = []
            expected = {}
            for keys, values in template.items():
                key = __get_key_part(keys, key_index)
                value = __get_value(values)
                required = __is_required_key(keys)
                param = __get_param_name_from_key(keys)
                if not required_only or required:
                    if isinstance(value, bool):
                        if value:
                            options.append(key)
                            options_str.append(key)
                    else:
                        options_str.append("{} {}".format(key, value))
                        options.append(key)
                        options.append(value)
                expected[param] = (
                    value if not required_only or required else __get_default(values)
                )
            yield {
                "options": options,
                "options_str": options_str,
                "input": None,
                "expected": expected,
            }


def __generate_testcases(options):
    for case in __generate_options_testcase(options):
        yield case
    for case in __generate_prompt_options_testcase(options):
        yield case


def __log_testcase(testcase, cmd):
    output = "\n"
    output += "Options:\n"
    for option in testcase["options_str"]:
        output += "  {}\n".format(option)
    if testcase["input"]:
        output += "Input:\n"
        for input in testcase["input"].split("\n"):
            output += "  {}\n".format(input)
    output += "Expected:\n"
    for key, value in testcase["expected"].items():
        if isinstance(value, str) and not value:
            value = '""'
        output += "  {}={}\n".format(key, value)
    logger.info(output)


def run_generate_command_test(runner, cmd, cmd_backend_run, options):
    # check missing options
    dynamic_params = [p.name for p in cmd.params if isinstance(p, DynamicOption)]
    options_params = [__get_param_name_from_key(key) for key in options]
    missing = []
    for name in dynamic_params:
        if name not in options_params:
            missing.append(name)
    assert not missing

    # 4 testcases
    # only required options
    # all options
    # only required prompt
    # all prompt
    for case in __generate_testcases(options):
        cmd_backend_run.reset_mock()
        __log_testcase(case, cmd)
        result = runner.invoke(cmd, case["options"], input=case["input"])
        assert result.exit_code == 0, result.output
        cmd_backend_run.assert_called_once()
        cmd_backend_run.call_args_list[0].kwargs == case["expected"]
    cmd_backend_run.reset_mock()



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.run.expo_app_publish import expo_app_publish as run_expo_app_publish
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)


ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["-l,--deployment-group"] = "group"
ALL_VALID_OPTIONS["!-u,--deployment-unit"] = "unit"
ALL_VALID_OPTIONS["-n,--node-package-manager"] = "auto"
ALL_VALID_OPTIONS["-o,--binary-output-dir"] = "binary_output_dir"


@mock.patch("hamlet.command.run.expo_app_publish.run_expo_app_publish_backend")
def test_input_valid(run_expo_app_publish_backend):
    run_options_test(
        CliRunner(),
        run_expo_app_publish,
        ALL_VALID_OPTIONS,
        run_expo_app_publish_backend.run,
    )


@mock.patch("hamlet.command.run.expo_app_publish.run_expo_app_publish_backend")
def test_input_validation(run_expo_app_publish_backend):
    runner = CliRunner()
    run_validatable_option_test(
        runner,
        run_expo_app_publish,
        run_expo_app_publish_backend.run,
        {
            "-u": "unit",
        },
        [],
    )



import os
import re
import click

from tabulate import tabulate

from hamlet.command import root as cli
from hamlet.command.common.display import json_or_table_option, wrap_text
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options
from hamlet.backend import query as query_backend
from hamlet.backend.create import template as create_template_backend


def find_schemas_from_options(options, schema):
    query_args = {
        **options.opts,
        "deployment_mode": None,
        "generation_entrance": "schemalist",
        "output_filename": "schemalist-schemacontract.json",
    }
    available_schemas = query_backend.run(
        **query_args,
        cwd=os.getcwd(),
        engine=options.engine,
        query_text=LIST_SCHEMAS_QUERY,
    )

    schemas = []

    for available_schema in available_schemas:
        if re.fullmatch(schema, available_schema["Schema"]):
            schemas.append(available_schema)
    return schemas


@cli.group("schema", context_settings=dict(max_content_width=240))
def group():
    """
    Generates JSONSchemas for Hamlet Configuration Scopes
    """
    pass


LIST_SCHEMAS_QUERY = "Stages[].Steps[]" ".{" "Schema:Parameters.Schema" "}"


def schema_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["Schema"]),
            ]
        )
    return tabulate(tablerows, headers=["Schema"], tablefmt="github")


@group.command(
    "list-schemas", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-s",
    "--schema",
    default=".*",
    show_default=True,
    help="A schema name pattern to filter results",
)
@json_or_table_option(schema_table)
@exceptions.backend_handler()
@pass_options
def list_schemas(options, schema):
    """
    List the available JSONSchemas that can be created
    """

    return find_schemas_from_options(options, schema)


@group.command(
    "create-schemas", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-s",
    "--schema",
    default=".*",
    show_default=True,
    help="A schema name pattern to filter the schemas that will be generated",
)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
    ),
    required=True,
    help="the directory where the outputs will be saved",
)
@exceptions.backend_handler()
@pass_options
def create_schemas(options, schema, output_dir, **kwargs):
    """
    Create the JSON schema files available
    """

    schema_list = find_schemas_from_options(options, schema)

    if len(schema_list) == 0:
        raise exceptions.CommandError("No schemas found")

    for schema_task in schema_list:
        click.echo("")
        click.echo(
            (
                click.style(
                    f"[*] Schema: {schema_task['Schema']}",
                    bold=True,
                    fg="green",
                )
            )
        )
        click.echo("")

        template_args = {
            **options.opts,
            "entrance": "schema",
            "output_dir": output_dir,
            "entrance_parameter": f"Schema={schema_task['Schema']}",
        }

        create_template_backend.run(
            **template_args, engine=options.engine, _is_cli=True
        )



import collections
from unittest import mock
from click.testing import CliRunner

from hamlet.command.release import update_image_reference as update_image_reference

from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)


ALL_VALID_OPTIONS = collections.OrderedDict()

ALL_VALID_OPTIONS["!-u,--deployment-unit"] = "deployment_unit"
ALL_VALID_OPTIONS["!-r,--build-reference"] = "build_reference"
ALL_VALID_OPTIONS["--code-tag"] = "code_tag"
ALL_VALID_OPTIONS["-f,--image-format"] = "docker"
ALL_VALID_OPTIONS["-s,--registry-scope"] = "registry_scope"


@mock.patch("hamlet.command.release.update_build_backend.UpdateBuildAutomationRunner")
def test_input_valid(update_image_reference_backend):
    run_options_test(
        CliRunner(),
        update_image_reference,
        ALL_VALID_OPTIONS,
        update_image_reference_backend,
    )


@mock.patch("hamlet.command.release.update_build_backend.UpdateBuildAutomationRunner")
def test_input_validation(update_image_reference_backend):
    runner = CliRunner()
    run_validatable_option_test(
        runner,
        update_image_reference,
        update_image_reference_backend,
        {
            "-u": "deployment_unit",
            "-r": "build_reference",
            "-f": "image_format",
        },
        [],
    )



import os

import boto3
import botocore.session
from botocore import credentials

# Use the cli cache so that we can run aws commands and boto commands with the same creds
cli_cache = os.path.join(os.path.expanduser("~"), ".aws/cli/cache")


def boto3_session(
    profile_name=None, config_file=None, credentials_file=None, existing_creds=None
):
    """
    returns a boto3 session using cached credentials
    Allows for sharing credentials from cli commands with boto sessions
    """
    # Construct botocore session with cache
    session = botocore.session.get_session()

    if profile_name is not None:
        session.set_config_variable("profile", profile_name)

    if config_file is not None:
        session.set_config_variable("config_file", config_file)

    if credentials_file is not None:
        session.set_config_variable("credentials_file", credentials_file)

    if existing_creds is not None:
        session.set_credentials(
            access_key=existing_creds["aws_access_key_id"],
            secret_key=existing_creds["aws_secret_access_key"],
            token=existing_creds["aws_session_token"],
        )

    session.get_component("credential_provider").get_provider(
        "assume-role"
    ).cache = credentials.JSONFileCache(cli_cache)

    return boto3.Session(botocore_session=session)



from hamlet.backend.common import runner


def run(
    account_repos=None,
    product_repos=None,
    commit_message=None,
    reference=None,
    tag=None,
    defer_push=None,
    log_level=None,
    engine=None,
    _is_cli=False,
    env={},
):

    opts = {
        "-a": account_repos,
        "-m": commit_message,
        "-p": product_repos,
        "-r": reference,
        "-t": tag,
    }
    env = {"AUTOMATION_LOG_LEVEL": log_level, "DEFER_REPO_PUSH": defer_push, **env}
    runner.run(
        "saveCMDBRepos.sh",
        args=[],
        options=opts,
        env=env,
        engine=engine,
        _is_cli=_is_cli,
        script_base_path_env="AUTOMATION_DIR",
    )



import subprocess
import shutil


def run(Command=None, env={}):
    """
    Run a bash command and use the stdout as the returned result
    """
    command_result = subprocess.check_output(
        [shutil.which("bash"), "-c", Command],
        text=True,
    )

    return {"Properties": {"result": command_result.rstrip("\n")}}



import os
from jinja2 import Template


class Group(object):
    """
    A cluster of entities
    """

    def __init__(self, dit):
        self.parentID = dit["parentID"]
        self.groupID = dit["groupID"]
        self.childIDs = []
        self.entities = []
        self.template = ""
        self.entityStr = ""
        self.rank = 0  # the rank of groups increment as it goes deeper
        self.libraries = []

    def findChildGroups(self, groups):
        """
        find child groups for parent groups, store childIDs for group object
        """
        for group in groups:
            if group.parentID == self.groupID and group != self:
                self.childIDs.append(group)

    def findEntities(self, entitiesGroup):
        """
        find entities of each group, store entities for group object
        """
        for en in entitiesGroup:
            if en["groupID"] == self.groupID:
                self.libraries.append(en["type"])  # save for library import
                en["type"] = en["type"].split(".")[-1]  # retrieve the last word as type
                self.entities.append(en)

    def createCluster(self):
        """
        create Cluster template
        """
        template = Template(
            """\
                with Cluster("{{groupID}}"):
                    {% for entity in entities %}
                    {{entity.entityID}}={{entity.type}}("{{entity.entityName}}")
                    {% endfor %}
            """
        ).render(groupID=self.groupID, entities=self.entities)
        self.entityStr = template
        self.template = template
        return self.template


def create_script(diagram, image_filename, temp_path):
    """
    Generates a diagrams formatted python script
    """

    diagramName = diagram["diagramName"]
    groups = diagram["groups"]
    entities = diagram["entities"]
    relationships = diagram["relationships"]

    image_filename = os.path.splitext(image_filename)[0]

    # build all group objects and store in list
    # create list of group objects, array in json is list in python
    groupObjects = []  # it has the length of group list
    for group in groups:
        groupObjects.append(Group(group))

    # find entities for each group and store
    for groupObj in groupObjects:
        groupObj.findEntities(entities)

    # match rank for each group and store
    rank_list = []

    def matchRank(groupObjects):
        allobjects = groupObjects.copy()
        if rank_list == []:
            for obj in groupObjects:
                # identify and remove groups with no parent
                if obj.parentID == "":
                    obj.rank = 1
                    rank_list.append(obj)
                    allobjects.remove(obj)
            groupObjects = allobjects
            matchRank(groupObjects)
        else:
            if allobjects != []:  # there are still objs unmatched
                for pa_obj in rank_list:
                    for obj in groupObjects:
                        if pa_obj.groupID == obj.parentID:
                            obj.rank = pa_obj.rank + 1
                            rank_list.append(obj)
                            allobjects.remove(obj)
                groupObjects = allobjects
                matchRank(groupObjects)
            else:
                return rank_list

    matchRank(groupObjects)

    # find child groups for each group and create template
    childGroups = []
    for groupObj in groupObjects:
        groupObj.findChildGroups(groupObjects)  # childIDs list is populated
        groupObj.createCluster()
        if (
            groupObj.childIDs == []
        ):  # if this group object has no child groups, save it as its own child group???
            childGroups.append(groupObj)

    def removeDuplicates(list):
        new_list = []
        for li in list:
            if li not in new_list:
                new_list.append(li)
        return new_list

    def findParents(childGroups, allObjects, rootGroups):
        """
        find parent group for each child group and combine Clusters
        """
        parentIDs = []
        for child in childGroups:
            if child.parentID == "":
                rootGroups.append(child)
            else:
                parentIDs.append(child.parentID)
        if parentIDs == []:
            return "all groups are root groups", removeDuplicates(rootGroups)
        parentIDs = removeDuplicates(parentIDs)

        parentsList = []
        for parent in parentIDs:
            template = Template(
                """\n
            with Cluster("{{parentID}}"):
                {% for child in childGroups %}
                {% if child.parentID == parentID %}
                {{child.template}}
                {% endif %}
                {% endfor %}
            """
            ).render(parentID=parent, childGroups=childGroups)
            for obj in allObjects:
                if obj.groupID == parent:
                    obj.template = obj.template + template
                    parentsList.append(obj)

        return findParents(parentsList, allObjects, rootGroups), removeDuplicates(
            rootGroups
        )

    # write all parent groups into template file
    with open(temp_path, "w+", encoding="utf-8") as f:
        for g in findParents(childGroups, groupObjects, [])[1]:
            f.write(g.template)

    # retrive parent groups file and remove redundant spaces. (Jinja2 template generates lots of spaces and empty lines)
    with open(temp_path, "r", encoding="utf-8") as f:
        lines = []
        for line in f.readlines():
            if line.strip() != "":
                lines.append(line)

    # insert remaining entities if there are any
    for obj in groupObjects:
        if obj.entities != [] and obj.childIDs != []:
            str = 'with Cluster("{}"):'.format(obj.groupID)
            lines = [obj.entityStr if l == str else l for l in lines]

    # write template in file
    with open(temp_path, "w+", encoding="utf-8") as f:
        for l in lines:
            f.write(l + "\n")

    # remove redundant spaces and empty lines
    with open(temp_path, "r", encoding="utf-8") as f:
        lines = f.readlines()
        lines = [l.strip() for l in lines]
        lines = [l for l in lines if l != ""]
        lines = removeDuplicates(lines)

        # get group rank for Cluster
        def clusterRank(str):
            name = str.split('"')[1].split('"')[0]
            for obj in groupObjects:
                if obj.groupID == name:
                    return obj.rank

        # import libraries from entity types
        import_diagrams = (
            "import diagrams\n"
            "from diagrams import Cluster, Diagram, Edge\n"
            "from diagrams.generic.blank import Blank\n"
        )
        imported_libraries = []
        for obj in groupObjects:
            for diagClass in obj.libraries:
                library = diagClass.rsplit(".", 1)[0]
                if library not in imported_libraries:
                    imported_libraries.append(library)
                    import_libray = "from {} import *\n".format(library)
                    import_diagrams = import_diagrams + import_libray

        exe_temp = (
            import_diagrams
            + f'\nwith Diagram("{diagramName}", show=False, outformat="png",'
            + f'filename="{image_filename}", direction="TB"):\n'
        )
        space = "    "  # indent
        rank = 0
        for i in range(len(lines)):
            if "with" in lines[i]:
                rank = clusterRank(lines[i])
                lines[i] = space * rank + lines[i] + "\n"  # indent according to ranks
            else:
                if "with" in lines[i - 1]:
                    rank = rank + 1
                lines[i] = space * rank + lines[i] + "\n"
            exe_temp = exe_temp + lines[i]

    # add relationships
    for rela in relationships:
        if isinstance(rela, dict):
            if rela["direction"] == "one way":
                rela_str = f'{space}{rela["startEntityID"]} >> Edge() >> {rela["endEntityID"]}\n'
                exe_temp = exe_temp + rela_str
            elif rela["direction"] == "two way":
                rela_str = f'{space}{rela["startEntityID"]} >> Edge() << {rela["endEntityID"]}\n'
                exe_temp = exe_temp + rela_str

    return exe_temp



import os
import json

from unittest import mock

from click.testing import CliRunner
from hamlet.command.visual import list_diagrams as list_diagrams
from hamlet.command.visual import list_diagram_types as list_diagram_types


def template_backend_run_mock(data):
    def run(output_filename="diagraminfo.json", output_dir=None, *args, **kwargs):
        os.makedirs(output_dir, exist_ok=True)
        info_filename = os.path.join(output_dir, output_filename)
        with open(info_filename, "wt+") as f:
            json.dump(data, f)

    return run


def mock_backend(info=None):
    def decorator(func):
        @mock.patch("hamlet.backend.query.template")
        def wrapper(blueprint_mock, *args, **kwargs):
            blueprint_mock.run.side_effect = template_backend_run_mock(info)
            return func(blueprint_mock, *args, **kwargs)

        return wrapper

    return decorator


@mock_backend(
    {
        "Diagrams": [
            {
                "Id": "DiagramId[1]",
                "Type": "DiagramType[1]",
                "Description": "DiagramDescription[1]",
            },
            {
                "Id": "DiagramId[2]",
                "Type": "DiagramType[2]",
                "Description": "DiagramDescription[2]",
            },
        ]
    }
)
def test_query_list_diagrams(blueprint_mock):
    cli = CliRunner()
    result = cli.invoke(list_diagrams, ["--output-format", "json"])
    print(result.exception)
    assert result.exit_code == 0
    result = json.loads(result.output)
    assert len(result) == 2
    assert {
        "Id": "DiagramId[1]",
        "Type": "DiagramType[1]",
        "Description": "DiagramDescription[1]",
    } in result
    assert {
        "Id": "DiagramId[2]",
        "Type": "DiagramType[2]",
        "Description": "DiagramDescription[2]",
    } in result


@mock_backend(
    {
        "DiagramTypes": [
            {
                "Type": "DiagramType[1]",
                "Description": "DiagramDescription[1]",
            },
            {
                "Type": "DiagramType[2]",
                "Description": "DiagramDescription[2]",
            },
        ]
    }
)
def test_query_list_diagram_types(blueprint_mock):
    cli = CliRunner()
    result = cli.invoke(list_diagram_types, ["--output-format", "json"])
    print(result.exception)
    assert result.exit_code == 0
    result = json.loads(result.output)
    assert len(result) == 2
    assert {"Type": "DiagramType[1]", "Description": "DiagramDescription[1]"} in result
    assert {"Type": "DiagramType[2]", "Description": "DiagramDescription[2]"} in result



import subprocess
import shutil
from pytest import ExitCode as ec

from hamlet.backend.common.exceptions import BackendException


def run(testpaths=None, root_dir=None, pytest_args=None, silent=True):
    testpaths = testpaths or []
    # Do not use pytest.main because it can't correctly work with the files changed in runtime

    if shutil.which("bash") is None:
        raise BackendException("Could not find bash installation")

    try:
        args = [
            "pytest",
            "-vvs" if not silent else "",
            "--cache-clear",
        ]

        if root_dir is not None:
            args.append(f"--rootdir={root_dir}")

        if pytest_args is not None:
            args.append(pytest_args)

        args += [
            *testpaths,
        ]

        cmd = " ".join(args)
        stdin = subprocess.PIPE
        stderr = subprocess.PIPE
        process = subprocess.Popen(
            [shutil.which("bash"), "-c", cmd],
            stdin=stdin,
            stderr=stderr,
            start_new_session=True,
            cwd=root_dir,
        )
        process.wait()
        if process.returncode in [ec.OK, ec.NO_TESTS_COLLECTED]:
            return True
        if process.returncode == ec.TESTS_FAILED:
            raise Exception("Tests Failed")
        if process.returncode in [ec.INTERNAL_ERROR, ec.USAGE_ERROR, ec.TESTS_FAILED]:
            raise Exception(process.stderr)
        else:
            raise Exception("Unknown exit code: %s" % process.returncode)
    finally:
        process.kill()
